"""Create a board object.

Create a board object from the board class to keep track of all parameters for the board.
The board object will hold the open connection to the hardware, this is useful since it
will make sure there is only one connection and it's through the board.

By keeping all the parameters stored with a board it open up to use multiple boards and
to switch board during operation.

Example:
--------

.. code-block:: python

    brd = Board('asocv3')
    brd.load_registers()
    brd.load_clockfile()
    brd.get_uart_connection(comport="COM99", baud=115200)

"""
import pathlib
from logging import getLogger
import naluconfigs
import naluconfigs.exceptions

from naludaq.backend.context import Context
from naludaq.backend.exceptions import BackendError
from naludaq.backend.managers.config import ConfigManager
from naludaq.backend.managers.connection import ConnectionManager
from naludaq.board.board_inits import BoardInits
from naludaq.board.connections import get_connection
from naludaq.board.params import (
    add_config_to_board_from_file,
    add_default_config_to_board,
)
from naludaq.communication.control_registers import ControlRegisters
from naludaq.controllers.board import get_board_controller
from naludaq.controllers.trigger import get_trigger_controller
from naludaq.controllers.trigger.default import BaseTriggerController
from naludaq.controllers.connection import get_connection_controller
from naludaq.controllers.external_dac import get_dac_controller
from naludaq.controllers.peripherals import get_peripherals_controller
from naludaq.controllers.readout import get_readout_controller
from naludaq.controllers.clock import get_clock_controller
from naludaq.helpers.exceptions import VersionError
from naludaq.helpers.helper_functions import get_available_port
from naludaq.helpers.validations import validate_dir_or_raise
from naludaq.io.io_manager import load_clockfile

LOGGER = getLogger(__name__)

try:
    import naludaq_rs
except ImportError:
    LOGGER.info("Could not import naludaq_rs, automatic start of backend not available")
    naludaq_rs = None


class Board:
    """Representation of the hardware.

    The board object is the software representation of the physical board.
    It should be created first since it's used by all other modules.
    The purpose of the board object is to mirror the hardware state.

    Once instantiated it holds a mirror of all the parameters on the physical unit,
    it holds the connection to the unit and stores the registers mapping.
    The connection and parameters can be accessed directly if needed:

        board.params is a dictionary with the board parameters.
            {param_name": value,
            ....
            }

        board.connection
            send() will send data to the board. Takes a 32-bit word (4 bytes or 8 hex chars).
            read() listens for data from the board.

        board.register is a dictionary
            {"register_type": {
                "name": [value, address, position, width, r/w],
                ...
            }}

    Example:
        ```
        brd = Board('asocv3');
        brd.get_uart_connection(comport="COM99", baud=115200);
        ```
        Will create a board object with the parameters for the asocv3 board.

    Args:
        model(str): model name of the board
        registers(Path): optional path to the registers file, loads default if not supplied.
        clock(Path): optional path to the clock file, loads default if not supplied.


    Raises:
        InvalidBoardModelError if the supplied board model doesn't exist.
    """

    def __init__(
        self,
        model: str = "default",
        registers: "str | pathlib.Path | None" = None,
        clock: "str | pathlib.Path | None" = None,
    ):
        self._is_reading_out: "bool | None" = False
        self.model: str = model
        self.server = None
        # Params are needed for default model
        self.params = {
            "model": model,
            "channels": 4,
        }
        self.features: dict = {}
        self.registers: dict = {}
        self.registers_cache: dict = {}

        # Connection
        self._using_new_backend: bool = False
        self._last_server_address: tuple[str, int] | None = None
        self.connection_info: "dict | None" = None
        self.connection = None
        self.spi_connection = None

        # Calibration
        self.pedestals: "dict | None" = None
        self.caldata: "dict | None" = None
        self.timingcal: "list | None" = None
        self.tuning: dict = {}

        self.trigger = None
        self.readout = None
        self.control = None
        self.ext_bias = None
        self.peripherals = None
        self.clock = None

        # Hardware
        if model not in ["default"]:
            self.load_registers(registers)
            self.load_clockfile(clock)

            self.trigger: BaseTriggerController = get_trigger_controller(self)
            self.readout = get_readout_controller(self)
            self.control = get_board_controller(self)
            self.ext_bias = get_dac_controller(self)
            self.peripherals = get_peripherals_controller(self)
            try:
                self.clock = get_clock_controller(self)
            except Exception:
                LOGGER.warning("No clock file found for %s", model)

    def __del__(self):
        """Cleanup the server if it's running."""
        self.stop_server()

    @property
    def context(self) -> "Context | None":
        """Get the backend context."""
        return getattr(self, "_context", None)

    @property
    def using_new_backend(self) -> bool:
        return self._using_new_backend

    @property
    def dac_values(self):
        """Get/Set the DAC values for the channels

        Simple pass thru to the correct spot in the params structure.
        Doesn't have any validation, is used both to set and get values.
        To set, simply run board.dac_values[chan] = value
        NO validations!
        """
        return self.params["ext_dac"]["channels"]

    @property
    def available_chips(self) -> int:
        """Get the number of chips on this board"""
        return len(self.params.get("chips", [0]))

    @property
    def selected_chips(self) -> list[int]:
        """Get/set the selected chips for this board.

        This is a pass through for the communication layer's select_chips / selected_chips functions.
        Selecting chips is done by toggling control registers to mask I/O from the chips that aren't selected.
        Make sure to restore the selected chips after you're finished with what you're doing.
        """
        from naludaq.communication import selected_chips

        return selected_chips(self)

    @selected_chips.setter
    def selected_chips(self, chips: list[int]):
        from naludaq.communication.chip_selection import select_chips

        select_chips(self, chips)

    @property
    def is_reading_out(self) -> bool:
        """Get boards readout status."""
        return self._is_reading_out

    @property
    def channels(self):
        """Get the channels from the params."""
        return self.params["channels"]

    def __enter__(self):
        """Connect to the board"""
        LOGGER.debug('Entering board context ("with" block); opening connection...')
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        # Handy debug logging to help users who break their code by using nested `with` blocks
        LOGGER.debug('Exiting board context ("with" block); closing connection...')
        self._in_context = False
        self.disconnect()

    def load_registers(self, registers_file=None, safe: bool = True):
        """Loads the configuration for the board, including:
        - `board.features`,
        - `board.params`,
        - and `board.registers`.

        If no file is supplied, then the defaults are used instead.

        Args:
            registers_file(str): Load from a YAML file, or from the defaults if None.
            safe (bool): whether to perform validation checks on the configuration.

        Raises:
            OSError: if the file could not be opened.
            InvalidParameterFile: if the configuration could not
                be loaded from the file or is otherwise invalid.
            InvalidRegisterFile: if the registers are invalid.
        """
        if registers_file is None:
            add_default_config_to_board(self, safe=safe)
        else:
            add_config_to_board_from_file(self, registers_file, safe=safe)

    def is_feature_enabled(self, name: str) -> bool:
        """Check whether the given feature is marked as enabled
        for this board.

        Args:
            name (str): the name of the feature (case sensitive).

        Returns:
            True if the feature is enabled, or False otherwise.
        """
        return self.features.get(name, False)

    def connect_serial(
        self, port: str, baud: int, backend_addr: tuple[str, int] = None
    ):
        """Connect to the board using a serial port. [BETA]

        Args:
            port (str): name of the comport (Windows) or device path (Linux)
            baud (int): preferred baud rate
            backend_addr: (tuple[str, int]): address of the backend,
                or None to use the previous backend
        """
        if baud is None:
            baud = max(self.params["possible_bauds"].keys())
        LOGGER.info(
            "Connecting serial with port=%s, baud_rate=%s, backend_addr=%s",
            port,
            baud,
            backend_addr,
        )
        self.connect_server(backend_addr)
        device = ConnectionManager(self).connect_serial(port, baud)
        self.connection_info = device.info
        get_connection_controller(self).configure_connection()

    def connect_d2xx(
        self, serial_number: str, baud: int, backend_addr: tuple[str, int] = None
    ):
        """Connect to the board using D2XX. [BETA]

        Args:
            serial_number (str): serial number of the board as a string
            baud (int): preferred baud rate
            backend_addr: (tuple[str, int]): address of the backend,
                or None to use the previous backend
        """
        if baud is None:
            baud = max(self.params["possible_bauds"].keys())
        LOGGER.info(
            "Connecting D2XX with serial_number=%s, baud_rate=%s, backend_addr=%s",
            serial_number,
            baud,
            backend_addr,
        )
        self.connect_server(backend_addr)
        device = ConnectionManager(self).connect_d2xx(serial_number, baud)
        self.connection_info = device.info
        get_connection_controller(self).configure_connection()

    def connect_d3xx(self, serial_number: str, backend_addr: tuple[str, int] = None):
        """Connect to the board using D3XX

        Args:
            serial_number (str): serial number of the board as a string
        """
        LOGGER.info("Connecting D3XX with serial_number=%s", serial_number)
        self.connect_server(backend_addr)
        device = ConnectionManager(self).connect_d3xx(serial_number)
        self.connection_info = device.info
        get_connection_controller(self).configure_connection()

    def connect_udp(
        self,
        board_addr: tuple[str, int],
        receiver_addr: tuple[str, int],
        backend_addr: tuple[str, int] = None,
    ):
        """Connect to the board using UDP. [BETA]

        Args:
            board_addr (tuple[str, int]): board socket address
            receiver_addr (tuple[str, int]): receiver socket address
            backend_addr: (tuple[str, int]): address of the backend,
                or None to use the previous backend
        """
        LOGGER.info(
            "Connecting UDP with board_addr=%s, receiver_addr=%s, backend_addr=%s",
            board_addr,
            receiver_addr,
            backend_addr,
        )
        self.connect_server(backend_addr)
        device = ConnectionManager(self).connect_udp(board_addr, receiver_addr)
        self.connection_info = device.info
        get_connection_controller(self).configure_connection()

    def _set_context_or_raise(self, backend_addr: tuple[str, int]):
        """Initialize a context for a backend with the given address.

        Args:
            backend_addr (tuple[str, int]): address of the backend.

        Raises:
            BackendError: if the server could not be reached
        """
        try:
            self._context = Context(backend_addr)
        except BackendError:
            raise
        self._last_server_address = backend_addr
        event_stop_word = self.params["stop_word"]
        answer_stop_word = self.params.get("register_stop_word", event_stop_word)
        ConfigManager(self).configure_packaging(
            events=event_stop_word, answers=answer_stop_word
        )
        self._using_new_backend = True

    def get_uart_connection(self, comport=None, serial_number=None, baud=None):
        """Connect to the board using UART.

        Test connecting either using comport or using serial
        sets up a connection as `self.connection`

        Args:
            comport (str): name of the comport
            serial_number(str): serial number of the board as a string
            baud (int): preferred baudrate
        """
        self.connection_info = {
            "type": "uart",
            "stop_word": self.params["stop_word"],
            "model": self.model,
            "usb_addr": comport,
            "speed": baud,
            "serial_number": serial_number,
            "new_firmware": self.params.get("new_firmware", False),
        }

        try:
            self.connection = get_connection(self.connection_info)
        except ConnectionError as e:
            raise ConnectionError(f"Failed to open connection: {e}") from e
        else:
            self._context = None
            self._using_new_backend = False
            get_connection_controller(self).configure_connection()

    def get_ftdi_connection(
        self, board_number=None, comport=None, serial_number=None, baud=None
    ):
        """Connect to the board using FTDI.

        Test connecting either using comport or using serial
        sets up a connection as `self.connection`

        Args:
            board_number (int): number of the FTDI index.
            serial_number(str): serial number of the board as a string
            baud (int): preferred baudrate
        """
        if baud is None:
            baud = self.params.get("default_baudrate", 2_000_000)

        self.connection_info = {
            "type": "ftdi",
            "stop_word": self.params["stop_word"],
            "model": self.model,
            "usb_addr": comport,
            "board_number": board_number,
            "speed": baud,
            "serial_number": serial_number,
        }

        try:
            self.connection = get_connection(self.connection_info)
        except ConnectionError as e:
            raise ConnectionError(f"Failed to open connection: {e}") from e
        else:
            self._context = None
            self._using_new_backend = False
            get_connection_controller(self).configure_connection()

    def get_ft60x_connection(
        self,
        *args,
        serial_number=None,
        **kwargs,
    ):
        """Connect to the board using FTDI.

        Test connecting either using comport or using serial
        sets up a connection as `self.connection`

        Args:
            board_number (int): number of the FTDI index.
            serial_number(str): serial number of the board as a string
            baud (int): preferred baudrate
        """
        self.connection_info = {
            "type": "ft60x",
            "stop_word": self.params["stop_word"],
            "model": self.model,
            "speed": 200_000_000,
            "serial_number": serial_number,
        }

        try:
            self.connection = get_connection(self.connection_info)
        except ConnectionError as e:
            raise ConnectionError(f"Failed to open connection: {e}") from e
        else:
            self._context = None
            self._using_new_backend = False
            get_connection_controller(self).configure_connection()

    def get_tcp_connection(self, ip: str, port: int):
        """Open a TCP connection."""
        self.connection_info = {
            "type": "tcp",
            "model": self.model,
            "stop_word": self.params["stop_word"],
            "ip": ip,
            "port": port,
            "speed": int(200e6),  # gigabit!
        }

        try:
            self.connection = get_connection(self.connection_info)
            self.connection.open()
        except ConnectionError as error_msg:
            self.connection = None
            raise ConnectionError("No connection established due to: %s", error_msg)
        else:
            self._context = None
            self._using_new_backend = False
            get_connection_controller(self).configure_connection()

    def get_udp_connection(self, board_addr: tuple, receiver_addr: tuple):
        """Open a UDP connection."""
        self.connection_info = {
            "type": "udp",
            "model": self.model,
            "stop_word": self.params["stop_word"],
            "board_addr": board_addr,
            "receiver_addr": receiver_addr,
            "speed": int(200e6),  # gigabit!
        }

        try:
            self.connection = get_connection(self.connection_info)
            self.connection.open()
        except ConnectionError as error_msg:
            self.connection = None
            raise ConnectionError("No connection established due to: %s", error_msg)
        else:
            self._context = None
            self._using_new_backend = False
            get_connection_controller(self).configure_connection()

    def get_mock_uart_connection(self, user_port, board_port, ip="127.0.0.1"):
        """Connect to a mock board.

        Args:
            user_port (tuple): the port of the user
            board_port (tuple): the port of the board
            ip (str): the ip address

        Raises:
            ConnectionError if a problem occurred
        """
        self.connection_info = {
            "ip": ip,
            "user_port": user_port,
            "board_port": board_port,
            "type": "uart",
            "mock": True,
            "stop_word": self.params["stop_word"],
            "model": self.model,
            "usb_addr": 0,
            "speed": 100,
        }

        try:
            self.connection = get_connection(self.connection_info)
        except ConnectionError as error_msg:
            raise ConnectionError("No connection established due to: %s", error_msg)
        else:
            self._context = None
            self._using_new_backend = False
            get_connection_controller(self).configure_connection()

    def load_clockfile(self, filepath=None):
        """
        Changes the clock_file parameter to filepath
        Loads the clock file into a list that is returned

        Args:
            filepath: full filepath to the clock file, or None to
                load the default clock file
        """
        if filepath is None:
            try:
                clock_data, clock_file = naluconfigs.get_clock(self.model)
            except Exception:
                raise
        else:
            try:
                if not pathlib.Path(filepath).exists():
                    raise FileNotFoundError(f"{filepath} not found")
            except Exception:
                raise
            try:
                clock_data = load_clockfile(filepath)
                clock_file = filepath

            except Exception:
                raise

        self.clock_data = clock_data
        self.params["clock_file"] = str(clock_file)
        # return self.params['clock_file']

    def connect(self):
        """Open the connection to the board. Must be called after
        using one of the ``get_XXX_connection()`` methods.

        Raises:
            ConnectionError: if the board could not be connected
        """
        if getattr(self, "connection", None) is None and self.connection_info is None:
            raise ConnectionError(
                "Cannot reconnect without a known connection. One of the get_XXX_connection() "
                "methods must be called before the connect() method. "
            )
        if self.is_connected:
            LOGGER.info("Already connected.")
            return

        if self.using_new_backend:
            ConnectionManager(self).connect_from_info(self.connection_info)
        else:
            self.connection.open()
        if not self.is_connected:
            raise ConnectionError(
                "Connection failed to open. The information used to connect to the board may "
                "be incorrect, or the board may be disconnected or powered off."
            )

    def disconnect(self):
        """Close the connection, leaving it ready for future use."""
        if self.using_new_backend:
            ConnectionManager(self).disconnect()
        else:
            try:
                self.connection.close()
            except Exception:  # pragma: no cover
                pass

    @property
    def is_connected(self) -> bool:
        """``True`` if the board is currently connected and accessible, or ``False`` otherwise."""
        connected = False
        if self.using_new_backend:
            try:
                connected = ConnectionManager(self).is_connected
            except Exception:
                pass
        else:
            connection = getattr(self, "connection", None)
            connected = connection is not None and connection.is_open
        return connected

    def connect_server(self, backend_addr: tuple[str, int] = None):
        """Connect to the backend server.

        Args:
            backend_addr (tuple[str, int]): the address of the backend server.
                If None, will attempt to connect to the previous server.

        Raises:
            BackendError: if the backend could not be reached, or attempted to
                reconnect without a prior connection.
        """
        if backend_addr is None and self._last_server_address is None:
            if naludaq_rs is None:
                raise BackendError(
                    "Cannot connect to server without the backend server started. Run `board.start_server()` first."
                )
        if backend_addr is None:
            backend_addr = self._last_server_address
        try:
            self._set_context_or_raise(backend_addr)
        except BackendError:
            raise
        try:
            self.connection_info = ConnectionManager(self).device.info
        except Exception:
            # not an error, just means it wasn't previously connected.
            # we don't really want to reset connection_info either,
            # since that means we can't use connect() to reconnect.
            pass

    def disconnect_server(self):
        """Disconnect from the server"""
        if self.using_new_backend:
            self._context = None

    @property
    def server_connected(self) -> bool:
        """Checks if the server is connected."""
        connected = False
        if self.context is not None:
            connected = self.context.is_alive
        return connected

    def start_server(
        self,
        output_dir: str,
    ):
        """Start the backend server.

        Args:
            backend_addr (tuple[str, int]): the address of the backend server.
            output_dir (str): the directory to save the output data.

        Raises:
            BackendError: if the backend could not be reached, or attempted to
                reconnect without a prior connection.
        """
        LOGGER.debug("Starting backend server")
        if naludaq_rs is None:
            raise BackendError(
                "Cannot start server without the backend server installed."
            )
        if self.server_connected:
            raise BackendError("Cannot start server while already connected.")
        validate_dir_or_raise(output_dir, "Output directory")

        port = get_available_port()
        backend_addr = ("127.0.0.1", port)

        try:
            self.server = naludaq_rs.Server(
                backend_addr,
                output_dir,
            )
            self.server.start()
        except Exception:
            self.stop_server()
            raise BackendError("Could not start backend server.")

        try:
            self.connect_server(backend_addr)
        except Exception:
            self.stop_server()
            raise BackendError("Could not connect to backend server.")

        LOGGER.debug("Backend server started using address: %s", backend_addr)

    def stop_server(self):
        """Stop the backend server."""
        if self.server:
            self.server.stop()
            self.server = None


def startup_board(board: Board, init_start=True, stat: list = []) -> bool:
    """Setup connection and transfer startup settings.

    The board needs to be booted to startup.
    Run the startup sequence to connect, validate connection and
    transfer all initial settings to the board.

    Args:
        board
        init_start: If False, load the init settings from the board else transfer from computer.

    Returns:
        True if board is ready to use, False if init failed.
    """
    # Determine if connection is UART
    if board.using_new_backend:
        is_uart = ConnectionManager(board).is_uart_based
    else:
        is_uart = board.connection.is_uart()

    if (
        board.params["model"] in ["upac32", "upaci", "zdigitizer"] and is_uart
    ):  # , "hdsocv1", "hdsocv1_evalr1", "hdsocv1_evalr2"]
        # init_board
        if init_hardware(board, init_start):
            status = "Hardware succesfully initialized."
            stat.append((100, status))
            LOGGER.debug(status)
        else:
            status = "Hardware NOT initialized."
            stat.append((0, status))
            LOGGER.debug(status)
            board.connection = None
            return False
        return True

    cc = get_connection_controller(board)

    # Match and Reset speed
    if is_uart:
        if cc.reset_connection():
            status = "Connection is reset to default."
            stat.append((20, "Resetting connection."))
            LOGGER.debug(status)
        else:
            status = "Connection failed to reset."
            stat.append((0, status))
            LOGGER.debug(status)
            return False

    # Verify connection?
    if cc.verify_connection():
        stat.append((40, "Verifying connection."))
        status = "Connection is verified and fully functional."
        LOGGER.debug(status)
    else:
        status = "Connection NOT verified."
        stat.append((0, status))
        LOGGER.debug(status)
        return False

    # init_board? is this really not what startup should be?
    if not init_hardware(board, init_start):
        status = "Hardware NOT initialized."
        stat.append((0, status))
        LOGGER.debug(status)
        board.connection = None
        return False
    else:
        status = "Hardware succesfully initialized."
        stat.append((60, status))
        LOGGER.debug(status)

    # set_speed
    if is_uart:
        try:
            status = "Changing speed"
            stat.append((80, status))
            cc.set_speed_or_revert_to_default()
        except ConnectionError as error_msg:
            status = f"Changing baudrate failed due to: {error_msg}"
            stat.append((0, status))
            LOGGER.error(status)
            board.connection = None

    # Connections shouldn't touch this flag.
    if board.is_connected:
        status = "Board successfully started."
        stat.append((100, status))
        return True

    board.connection = None
    return False


def init_hardware(board: Board, init_start=True):
    """Try to init the hardware.

    Every board has a specific set of instructions to start, initialize and
    configure the board. This runs the init function for the board in use
    and without the need to specify the board model.

    Args:
        board (board): The board object representing the physical board to
        initialize.
        init_start(bool): Init board if True or read values from hw if False.

    Returns:
        True if successfully initialized.
    """
    # Determine if connection is UART
    if board.using_new_backend:
        is_uart = ConnectionManager(board).is_uart_based
    else:
        is_uart = board.connection.is_uart()

    # Init should run at lowest speed.
    if is_uart:
        get_connection_controller(board).reset_speed()

    _read_firmware_version(board)
    _validate_firmware_version_or_raise(board)

    if init_start:
        LOGGER.debug("Trying to init board...")
        if not _init_board(board):
            LOGGER.error("Board init failed...")
            return False
        else:
            LOGGER.debug("Board successfully initalized.")
    else:
        ControlRegisters(board).read_all(True)

    if is_uart:
        get_connection_controller(board).set_speed_or_revert_to_default()

    return True


def _read_firmware_version(board):
    """Update the firmware version on the software board with the hardware firmware number."""
    try:
        board.params["firmware_version"] = get_board_controller(
            board
        ).read_firmware_version()
    except Exception as error_msg:
        LOGGER.error("Failed to read firmware version due to: %s", error_msg)
        board.params["firmware_version"] = "Unknown"


def _validate_firmware_version_or_raise(board: Board):
    """Raise an error if the firmware version on the board is too old to
    be compatible with NaluDAQ.

    Uses the "minimum_firmware_version" field in the params. If this field
    is not present, or the firmware version was not determined, no
    error is raised.

    Args:
        board (Board): board object

    Raises:
        VersionError: if the firmware version is too old.
    """
    version = board.params.get("firmware_version", None)
    minimum_version = board.params.get("minimum_firmware_version", None)
    can_compare = None not in (version, minimum_version) and version > 0
    if can_compare and version < minimum_version:
        raise VersionError(
            "The board is programmed with a deprecated firmware version.\n"
            f"The current firmware version is v{version}, while the minimum\n"
            f"required version is v{minimum_version}.\n\n"
            "Please update the firmware or use an older software version."
        )


def _init_board(board) -> bool:
    """This will run the boardspecific initialization.

    The BoardInits is a factory getting the board init sequence based on the model.
    """
    return BoardInits(board).run_init_function()



import logging
import time

from naludaq.communication import AnalogRegisters, ControlRegisters, DigitalRegisters

from .default import BoardController

logger = logging.getLogger("naludaq.board_controller.hiper")


class BoardControllerHiper(BoardController):
    def read_scalar(self, channel: int):
        raise NotImplementedError("Not yet supported")

    def reset_board(self):
        self._write_control_register("txin_en", 0)
        self._write_control_register("rxout_en", 0)
        time.sleep(0.02)
        self.sysrst()
        time.sleep(0.02)
        self._write_control_register("rxout_en", self.chip_mask())

    def chip_mask(self, chips=None):
        if chips is None:
            chips = self.board.params["installed_chips"]
        return sum([1 << chip for chip in chips])

    def start_readout(
        self,
        trig="ext",
        lb="trigrel",
        acq="raw",  # "pedsub",
        dig_head=False,
        ped="zero",
        readoutEn=True,
        singleEv=False,
    ):
        mask = self.chip_mask()
        logger.debug("Start readout of chips: %s", self.board.params["installed_chips"])
        self._write_control_register("rxout_en", mask)

        # self._write_digital_register("miscreg", 2)
        # self._write_digital_register("pllmisc", 0)
        # self._write_digital_register("txspeed", 2)
        # self._write_digital_register("waitread", 1)

        self._write_control_register("exttrig_en", mask)
        self._write_control_register("txin_en", mask)

        super().start_readout(
            trig=trig,
            lb=lb,
            acq=acq,
            dig_head=dig_head,
            ped=ped,
            readoutEn=readoutEn,
            singleEv=singleEv,
        )

    def _generate_start_readout_cmd(self, readout_params, *args, **kwargs):
        (
            trigger_type,
            lbType,
            acq_type,
            pedestals_type,
        ) = self._convert_readout_params_to_binary(readout_params)

        readoutEn = readout_params.get("readoutEn", 1)
        singleEv = readout_params.get("singleEv", 0)
        dig_head = readout_params.get("dig_head", 0)

        first_part = hex(
            int(str(int(singleEv)) + str(int(readoutEn)) + pedestals_type + "0011", 2)
        )[2:].zfill(2)

        second_part = hex(
            int("000000" + trigger_type + str(int(dig_head)) + acq_type + lbType, 2)
        )[2:].zfill(3)

        return f"B{first_part}00{second_part}"

    def stop_readout(self):
        """Upated stop command.

        All Txin_en OFF
        All Rxout_en OFF
        ASIC Sysrst
        ASIC RegClr
        All RxIn_en ON
        Verify Busy_locked
        Load default ASIC Digital Registers
        Load default ASIC Analog Registers.
        """
        self.is_reading_out = False
        cmd = "B0B00000"
        self._send_command(cmd)
        self._write_control_register("rxout_en", self.chip_mask())
        self._write_control_register("txin_en", 0)

    def toggle_trigger(self):
        self.arm_trigger()
        cmd = "AF040440" + "AF040040"
        self._send_command(cmd)

    def regclr(self):
        self._write_control_register("regclr", True)
        self._write_control_register("regclr", False)

    def _sync_tx_pol(self):
        """Set ASIC txin polarity and check if locked

        SET POLARITY TO ASIC TXIN
        IF POLAROITY AND BAUD IS NOT CORRECT IDLE_DETECT WILL BE 0
        IF POLARITY AND BAUD ARE CORRECT IDLE_DETETCT WILL BE 1
        """
        for _ in range(5):
            # ControlRegisters(self.board).write('txin_pol', 0)  # 1900 hex
            if ControlRegisters(self.board).read("idle_det")["value"] == 1:
                return True
            time.sleep(0.05)
        return False

    def _sync_rx_pol(self):
        """Set ASIC rxout pol and make sure it's locked

        SET POLARITY TO ASIC RXOUT
        IF POLAROITY AND BAUD IS NOT CORRECT BUSY_LOCKED WILL BE 0
        IF POLARITY AND BAUD ARE CORRECT BUSY_LOCKED WILL BE 1
        """
        # ControlRegisters(self.board).write("txin_en", self.chip_mask())
        for _ in range(2):
            # ControlRegisters(self.board).write('rxout_pol', 0)  # 113C hex
            busy_locked = ControlRegisters(self.board).read("busy_locked")["value"]
            if busy_locked == 1:
                return True
            logger.debug("Busy locked returned: %s", busy_locked)
            time.sleep(0.05)
        return False

    def get_available_chips(self):
        """Get a list of available chip numbers."""
        idl = ControlRegisters(self.board).read("idle_det")["value"]
        idl_str = str(bin(idl))[2:]
        available_chips = []
        for chip_num, i in enumerate(range(len(idl_str), 0, -1)):
            if int(idl_str[i - 1]):
                available_chips.append(chip_num)

        return available_chips

    def arm_trigger(self):
        """Reset the trigger counter, allowing for more triggered readouts."""
        ControlRegisters(self.board).write("trig_count_reset", True)
        ControlRegisters(self.board).write("trig_count_reset", False)

    def _write_digital_register(self, name: str, value: int):
        DigitalRegisters(self.board).write(name, value)

    def set_loopback_enabled(self, enabled: bool):
        """Set whether serial loopback is enabled.

        Loopback can safely be disabled during most of the operations with the board.
        Loopback **must** be disabled when communicating over the serial interface.
        If serial communication with the ASIC is intended then this should run during startup and only be enabled as needed.

        Args:
            enabled (bool): True to enable loopback.

        Raises:
            TypeError if enabled is not a bool.
        """
        if not isinstance(enabled, bool):
            raise TypeError("Argument must be bool")
        OFF = "B0900002"
        ON = "B0900003"
        cmd = ON if enabled else OFF
        self._send_command(cmd)

    def enable_chip(self, chips: "list[int] | int"):
        """Enable the chip for readout.

        Will turn on the chip from a disabled state.
        Only turn on the one chip that is needed for readout.

        Args:
            chip (int): The chip number to enable.
        """
        # mask = self.chip_mask([chip])
        # self._write_control_register("rxout_en", mask)
        if isinstance(chips, int):
            chips = [chips]
        self.analog_startup(chips)
        self.digital_startup(chips)
        self._write_control_register("rxout_en", self.chip_mask())

    def disable_chip(self, chips: "list[int] | int"):
        """Disable the chip for readout.

        Will turn off the chip and it will not work properly without re-enabling it.

        Args:
            chip (int): The chip number to disable.
        """
        if isinstance(chips, int):
            chips = [chips]
        mask = self.chip_mask(chips)
        ControlRegisters(self.board).write("chip_sysrst", mask)
        ControlRegisters(self.board).write("chip_sysrst", 0)
        ControlRegisters(self.board).write("chip_regclr", mask)
        ControlRegisters(self.board).write("chip_regclr", 0)
        self._write_control_register("rxout_en", self.chip_mask())

    # def write_commands(self, commands):
    #     for cmd in commands:
    #         self.board.connection.send(cmd)
    #         time.sleep(0.01)

    def digital_startup(self, chips: "list[int] | int | None" = None):
        """Start the digital side of the chip by programming all registers."""
        DigitalRegisters(self.board, chips=chips).write_all()

    def analog_startup(self, chips: "list[int] | int | None" = None):
        """Start the analog side of the chip by programming all registers."""
        AnalogRegisters(self.board, chips=chips).write_all()
        time.sleep(0.1)
        self.dll_startup(chips=chips)

    def dll_startup(self, chips: "list[int] | int | None" = None):
        """Starting the delay line on the analog side.

        Sets and unsets vanbuff to get the dll going and
        changes the vadjp values to ensure proper SST duty cycle once locked.
        """
        AnalogRegisters(self.board, chips=chips).write("qbias", 0)
        AnalogRegisters(self.board, chips=chips).write("vadjn_sw", True)
        time.sleep(1)
        AnalogRegisters(self.board, chips=chips).write("qbias", 2048)
        AnalogRegisters(self.board, chips=chips).write("vadjn_sw", False)



"""
"""
from logging import getLogger
from naludaq.controllers.trigger.hdsoc import TriggerControllerHdsoc
from naludaq.helpers.helper_functions import type_name
from enum import Enum

logger = getLogger("naludaq.trigger_controller_hdsoc")

# {TIA_en, TIA_DC_adj[0:6], tsel, ch_en, scv_legacy, tsgn}
TSEL_BIT = 3
TSGN_BIT = 0


class TriggerGroup(Enum):
    """Used to select grouped channels for setting trigger thresholds and references."""

    CH0_15 = 0
    CH16_31 = 1
    CH32_47 = 2
    CH48_63 = 3


class TriggerControllerHdsocv2(TriggerControllerHdsoc):
    def __init__(self, board):
        super().__init__(board)

    @property
    def references(self) -> "dict[str, tuple[int, int]]":
        """Get/Set the current low/high references for the left and right side

        - nb, left: 16_31
        - nb, right: 48_63
        - sb, left: 0_15
        - sb, right: 32_47

        Returns:
            dict[str, tuple[int, int]]: A dictionary where the keys are the side
                                        and the values are tuples of low and high references.
        """

        ref0_15_low = self._get_analog_register("sub_ref_neg_0_15_trigger")
        ref0_15_high = self._get_analog_register("sub_ref_pos_0_15_trigger")
        ref16_31_low = self._get_analog_register("sub_ref_neg_16_31_trigger")
        ref16_31_high = self._get_analog_register("sub_ref_pos_16_31_trigger")
        ref32_47_low = self._get_analog_register("sub_ref_neg_32_47_trigger")
        ref32_47_high = self._get_analog_register("sub_ref_pos_32_47_trigger")
        ref48_63_low = self._get_analog_register("sub_ref_neg_48_63_trigger")
        ref48_63_high = self._get_analog_register("sub_ref_pos_48_63_trigger")

        ref = {
            "0_15": (ref0_15_low, ref0_15_high),
            "16_31": (ref16_31_low, ref16_31_high),
            "32_47": (ref32_47_low, ref32_47_high),
            "48_63": (ref48_63_low, ref48_63_high),
        }
        return ref

    @references.setter
    def references(self, values: "dict[str, tuple[int, int]]"):
        """Set the low/high references for the left and right side"""
        self._set_references(values)

    @property
    def low_references(self) -> dict[str, int]:
        """Get the current low references for the left and right side"""
        ref0_15_low = self._get_analog_register("sub_ref_neg_0_15_trigger")
        ref16_31_low = self._get_analog_register("sub_ref_neg_16_31_trigger")
        ref32_47_low = self._get_analog_register("sub_ref_neg_32_47_trigger")
        ref48_63_low = self._get_analog_register("sub_ref_neg_48_63_trigger")
        refs = {
            "0_15": ref0_15_low,
            "16_31": ref16_31_low,
            "32_47": ref32_47_low,
            "48_63": ref48_63_low,
        }
        return refs

    @low_references.setter
    def low_references(self, values: "dict[str | str, int]"):
        """Set the low references for the left and right side"""
        if not isinstance(values, dict):
            raise TypeError(f"Values must be a dict, not {type_name(values)}")
        for side, value in values.items():
            if side in [0, "0_15", TriggerGroup.CH0_15]:
                regname = "sub_ref_neg_0_15_trigger"
            elif side in [1, "16_31", TriggerGroup.CH16_31]:
                regname = "sub_ref_neg_16_31_trigger"
            elif side in [2, "32_47", TriggerGroup.CH32_47]:
                regname = "sub_ref_neg_32_47_trigger"
            elif side in [3, "48_63", TriggerGroup.CH48_63]:
                regname = "sub_ref_neg_48_63_trigger"
            else:
                raise ValueError(f"Invalid side: {side}")
            self._write_analog_register(regname, value)

    @property
    def high_references(self) -> dict[str, int]:
        """Get the current high references for the channel groups"""
        ref0_15_high = self._get_analog_register("sub_ref_pos_0_15_trigger")
        ref16_31_high = self._get_analog_register("sub_ref_pos_16_31_trigger")
        ref32_47_high = self._get_analog_register("sub_ref_pos_32_47_trigger")
        ref48_63_high = self._get_analog_register("sub_ref_pos_48_63_trigger")
        refs = {
            "0_15": ref0_15_high,
            "16_31": ref16_31_high,
            "32_47": ref32_47_high,
            "48_63": ref48_63_high,
        }
        return refs

    @high_references.setter
    def high_references(self, values: "dict[int | str, int]"):
        """Set the high references for the left and right side"""
        if not isinstance(values, dict):
            raise TypeError(f"Values must be a dict, not {type_name(values)}")
        for side, value in values.items():
            if side in [0, "0_15", TriggerGroup.CH0_15]:
                regname = "sub_ref_pos_0_15_trigger"
            elif side in [1, "16_31", TriggerGroup.CH16_31]:
                regname = "sub_ref_pos_16_31_trigger"
            elif side in [2, "32_47", TriggerGroup.CH32_47]:
                regname = "sub_ref_pos_32_47_trigger"
            elif side in [3, "48_63", TriggerGroup.CH48_63]:
                regname = "sub_ref_pos_48_63_trigger"
            else:
                raise ValueError(f"Invalid side: {side}")
            self._write_analog_register(regname, value)

    @property
    def edge(self) -> bool:
        """Get the trigger edge for all channels on the board."""
        raise NotImplementedError("Single edge is not implemented for HDSOCv2")

    @edge.setter
    def edge(self, value: bool):
        """Set the trigger edge for all channels on the board."""
        if not isinstance(value, bool):
            raise TypeError("edge must be a bool")
        self.edges = value

    @property
    def edges(self) -> "dict[int | str | TriggerGroup, bool]":
        """Get the trigger edge for all channels on the board."""
        edges = {}
        for ch in range(self.board.channels):
            reg = self._get_fwd_reg_name(ch)
            edges[ch] = bool(
                self.board.registers["analog_registers"][reg]["value"][0] & TSGN_BIT
            )
        return edges

    @edges.setter
    def edges(self, value: "dict[int | str | TriggerGroup, bool] | bool"):
        """Set the trigger edge for all channels on the board."""
        if isinstance(value, bool):
            edges = {ch: value for ch in range(self.board.channels)}
        elif isinstance(value, dict):
            edges = value
        else:
            raise TypeError("edges must be a dict or bool")
        self._set_trigger_edges(edges)

    @property
    def tsel(self) -> dict[int, bool]:
        """Get the trigger select for all channels on the board."""
        tsel = {}
        for ch in range(self.board.channels):
            reg = self._get_fwd_reg_name(ch)
            tsel[ch] = bool(
                self.board.registers["analog_registers"][reg]["value"][0] & TSEL_BIT
            )
        return tsel

    @tsel.setter
    def tsel(self, value: "bool | dict[int, bool]"):
        """Set the trigger select for all channels on the board."""
        if isinstance(value, bool):
            tsel = {ch: value for ch in range(self.board.channels)}
        elif isinstance(value, dict):
            tsel = value
        else:
            raise TypeError("channels must be a dict")

        self._update_fwd_regs(tsel, TSEL_BIT)

    def _set_references(self, values: dict[str, tuple[int, int]]):
        """Set the low/high references for the left and right side
        Args:
            values (dict[str, tuple[int, int]]): A dictionary where the keys are the side
                                                and the values are tuples of low and high references.
        Raises:
            TypeError: If the values argument is not a dictionary.
            TypeError: If the low or high reference is not an integer.
            ValueError: If the side is not valid.
        """
        if not isinstance(values, dict):
            raise TypeError("values must be a dict")
        for side, (low, high) in values.items():
            if not isinstance(low, int) or not isinstance(high, int):
                raise TypeError("Low and high references must be integers")
            if side == "0_15":
                self._write_analog_register("sub_ref_neg_0_15_trigger", low)
                self._write_analog_register("sub_ref_pos_0_15_trigger", high)
            elif side == "16_31":
                self._write_analog_register("sub_ref_neg_16_31_trigger", low)
                self._write_analog_register("sub_ref_pos_16_31_trigger", high)
            elif side == "32_47":
                self._write_analog_register("sub_ref_neg_32_47_trigger", low)
                self._write_analog_register("sub_ref_pos_32_47_trigger", high)
            elif side == "48_63":
                self._write_analog_register("sub_ref_neg_48_63_trigger", low)
                self._write_analog_register("sub_ref_pos_48_63_trigger", high)
            else:
                raise ValueError(f"Invalid side: {side}")

    def _set_tsel_enabled(self):
        """Enables TSEL (Trigger Select) for all channels on the board.

        This method creates a dictionary where each channel on the board is set to True,
        indicating that TSEL is enabled for that channel. It then calls the `enable_tsel`
        method with this dictionary to apply the settings.
        """
        channels = {i: True for i in range(self.board.channels)}
        self.tsel = channels

    def _set_trigger_edges(self, channels: dict[int, bool]):
        """Set trigger edge per channel, either rising=True, or rising=False

        Args:
            channels (dict[int, bool]): keys are channel numbers,
                                        value are True (rising)
                                        or false (falling).

        Raises:
            TypeError if channels is not a dict.
        """
        if not isinstance(channels, dict):
            raise TypeError("channels must be a dict")

        self._update_fwd_regs(channels, TSGN_BIT)

    def _set_trigger_thresholds(self, values: dict[int, int]):
        """Set the trigger values for the individual channels using the values
        stored in the board object.
        """
        values = values or self.values
        self._validate_thresholds_or_raise(values)
        self._set_tsel_enabled()
        for channel, threshold_value in values.items():
            register_name = self._trig_thresh_regname.format(channel)
            if (
                self.board.registers["analog_registers"][register_name]["value"][0]
                != threshold_value
            ):
                self._write_analog_register(register_name, threshold_value)

    def _update_fwd_regs(self, channels: dict[int, bool], bitpos: int):
        """Updates the forward registers for the specified channels.

        This method iterates over the provided channels dictionary, retrieves the
        corresponding register name for each channel, and sets a single bit at the
        specified bit position in the register based on the boolean value.

        Args:
            channels (dict[int, bool]): A dictionary where the keys are channel
                                        numbers and the values are
                                        booleans indicating the state.
            bitpos (int): The bit position to update in the register.
        """
        for ch, val in channels.items():
            reg = self._get_fwd_reg_name(ch)
            self._set_addr_single_bit(reg, bitpos, val)

    def _get_fwd_reg_name(self, ch: int) -> int:
        """Generate the forward register name for a given channel.

        Args:
            ch (int): The channel number.

        Returns:
            int: The formatted forward register name as a string.
        """
        regname = "ch_fwd_{}"
        reg = regname.format(ch)
        return reg

    def _set_addr_single_bit(self, reg: str, bitpos: int, bitval: bool):
        """Sets a single bit in a specified analog register.

        Will only send the write command if the bit value is different from the current value.

        Args:
            reg (str): The name of the register to modify.
            bitpos (int): The position of the bit to set.
            bitval (bool): The value to set the bit to (True for 1, False for 0).

        Returns:
            None
        """
        regval = self.board.registers["analog_registers"][reg]["value"]
        if isinstance(regval, list):
            regval = regval[0]
        rval = bit_replace(regval, bitpos, bitval)
        if rval != regval:
            self._write_analog_register(reg, rval)

    def _set_wbiases(self):
        """Enable the biasing system if it's not already enabled"""
        pass


def bit_replace(num: int, pos: int, val: int) -> int:
    """Replaces the bit at a specified position in an integer with a given value.

    Args:
        num (int): The original integer.
        pos (int): The position of the bit to replace (0-indexed from the right).
        val (int): The new value for the bit (0 or 1).

    Returns:
        int: The modified integer with the bit at the specified position replaced by the given value.
    """
    mask = ~(1 << pos)
    num &= mask
    num |= val << pos
    return num



from naludaq.backend.client import HttpClient
from naludaq.backend.exceptions import BackendError

DEFAULT_LOCAL_CONTEXT_ADDRESS = ("127.0.0.1", 7878)


class Context:
    def __init__(self, addr: tuple[str, int] = None):
        """Utility for communicating with the Rust backend.

        Args:
            addr (tuple[str, int]): address of the backend. Defaults to
                the default address for a local backend.

        Raises:
            HttpError: if the backend is unreachable.
        """
        if addr is None:
            addr = DEFAULT_LOCAL_CONTEXT_ADDRESS
        self._client = HttpClient(addr)
        if not self._client.reachable():
            raise BackendError(f"The server at {addr} is unreachable")

    def __eq__(self, other) -> bool:
        """Check if this context is equal to another.

        Only checks the address of the two contexts.
        """
        return isinstance(other, Context) and self.address == other.address

    @property
    def address(self) -> tuple[str, int]:
        """The host which the backend is running on"""
        return self._client.address

    @property
    def client(self) -> HttpClient:
        """The client used to communicate with the backend."""
        return self._client

    @property
    def is_alive(self) -> bool:
        """Tries to ping the server to check if it's reachable."""
        return self._client.reachable()



"""Digital Registers
====================
Registers controlling the digital part of the ASIC.

This is reliant on the digital_registers portion of the .yml file.

These registers control the digital side of the ASICs.
By changing these registers it's possible to
change how and what the ASIC is reading.

Command structure:
------------------
Updating a digital command uses a format 'B01AAVVV'

- B tells the FPGA command parser to expect a analog/digital command.
- M is mode, 0 = write, 1 = read
- AA = address
- VVV = value

Will be interpereted by the FPGA as hex. it's 32-bit = 8 hex values.

Example:
---------

.. code-block: python

    dc = DigitalRegisters(board)
    dc.read("regname")
    dc.write("regname", intvalue)

"""
from logging import getLogger

from naludaq.communication import _chip

LOGGER = getLogger(__name__)


class DigitalRegisters(_chip.ChipRegisters):
    def __init__(self, board, chips: "int | list[int]" = None):
        """Registers controlling the digital part of the ASIC(s).

        For a more thorough explanation of the multichip aspect and its implications,
        see the `multichip` module.

        Args:
            board (Board): board object
            chips (int | list[int] | None): chip(s) to read/write. If None, all chips are assumed.
                Only one chip may be read from at a time!
        """
        super().__init__(board, "digital_registers", chips)
        self._write_cmd = 0xB01
        self._read_cmd = 0xB42
        self._width_cmd = 3
        self._width_addr = 2  # hex char
        self._width_val = 3  # hex char

    @_chip.raise_if_multiple_chips
    def read_all(self):
        """Read all registers.
        Reimplemented to bypass the new backend way of doing this.

        Returns:
            dict: register values structured as {reg_name: reg dict}
        """
        # Digital registers have no command ID, so we need to read them synchronously
        values = {
            name: register
            for addr in self.list_addresses()
            for name, register in self.read_addr_named(addr).items()
        }
        return values

    @_chip.raise_if_multiple_chips
    def read_many(self, names: list[str]) -> list[str]:
        """Read several registers at once.
        Reimplemented to bypass the new backend way of doing this.

        Args:
            names (Iterable[str]): names of registers to read

        Returns:
            list[int]: list of register values in the same order as provided.
        """
        self._validate_names_or_raise(names)

        # Digital registers don't have command IDs so concurrent reads are dangerous
        return [self.read(name)["value"] for name in names]



import logging
import time
from functools import partial

from naludaq.devices import LTC2990
from naludaq.helpers import type_name
from naludaq.helpers.exceptions import BadDataError

from .peripherals_controller import PeripheralsController

logger = logging.getLogger("naludaq.peripherals_hdsoc")


class HdsocPeripheralsController(PeripheralsController):
    def __init__(self, board):
        """HDSoCv1 peripherals controller.

        Provides a way to read various I2C devices on the board.

        Args:
            board (Board): the board object.
        """
        super().__init__(board)
        self._reader_functions = {
            "temp": self.read_temp_of_chip,
            "current": self.read_current_of_chip,
            "fpga_voltage": self.read_fpga_voltage,
            "board_voltage": self.read_board_current,
            "vadjn_left": partial(self.read_vadjn, "left"),
            "vadjn_right": partial(self.read_vadjn, "right"),
            "vadjp_left": partial(self.read_vadjp, "left"),
            "vadjp_right": partial(self.read_vadjp, "right"),
        }

    def read_vadjn(self, side: str) -> float:
        """Reads a VadjN measurement from the board.

        Args:
            side (str): the side of the chip to measure,
                'left' or 'right'

        Returns:
            The measurement in Volts.
        """
        self._validate_side_or_raise(side)

        voltage = self._request_readout(f"vadjn_{side.lower()}")
        logger.debug("VadjN (%s): %s mV", side, voltage)
        return voltage

    def read_vadjp(self, side: str) -> float:
        """Reads a VadjP measurement from the board.

        Args:
            side (str): the side of the chip to measure,
                'left' or 'right'

        Returns:
            The measurement in Volts.
        """
        self._validate_side_or_raise(side)
        voltage = self._request_readout(f"vadjp_{side.lower()}")
        logger.debug("VadjP (%s): %s mV", side, voltage)
        return voltage

    def read_fpga_voltage(self) -> float:
        """Reads the FPGA supply voltage.

        Raises:
            BadDataError if the voltage cannot be read.
        """
        return self._read_ltc2990_voltage("fpga_voltage")

    def read_chip_voltage(self) -> float:
        """Reads the chip voltage in Volts.

        Raises:
            BadDataError if the voltage cannot be read.
        """
        return self._read_ltc2990_voltage("chip_voltage")

    def read_board_voltage(self) -> float:
        """Reads the board supply voltage in Volts.

        Raises:
            BadDataError if the voltage cannot be read.
        """
        return self._read_ltc2990_voltage("board_voltage")

    def read_board_current(self) -> float:
        """Calculates the board current draw in Amps.

        Raises:
            BadDataError if the voltage cannot be read.
        """
        fpga_voltage = self.read_fpga_voltage()
        board_voltage = self.read_board_voltage()
        resistor = self.board.params["peripherals"].get("current_resistor", 5e-3)
        return (fpga_voltage - board_voltage) / resistor

    def _read_ltc2990_voltage(self, param_name: str, num_samples=1):
        """Reads a voltage from the LTC2990.

        Args:
            param_name (str): the name of the param in the board YAML.
            num_samples (int): number of sample points to average.
                Measurements jitter a bunch, so averaging is helpful.

        Returns:
            The voltage in Volts.

        Raises:
            BadDataError if the voltage cannot be read.
            NotImplementedError if the given parameter name is invalid.
        """
        try:
            params = self._get_peripheral_params(param_name)
            addr, chan = params["addr"], params["chan"]
        except (KeyError, NotImplementedError):
            raise NotImplementedError(
                f'Peripheral measurement "{param_name}" is not defined'
            )

        monitor = LTC2990(self._board, addr)

        samples = []
        for _ in range(num_samples):
            monitor.trigger_conversion()  # Important! Conversion needed before reading
            time.sleep(0.01)
            if chan == "vcc":
                try:
                    voltage = monitor.read_vcc()
                except BadDataError:
                    raise
            else:
                try:
                    voltage = monitor.read_voltage(chan)
                except BadDataError:
                    raise

            samples.append(voltage)

        return sum(samples) / len(samples)

    @staticmethod
    def _validate_side_or_raise(side: str):
        """Checks whether a given side value is 'left' or 'right'.
        Not case sensitive.

        Raises:
            TypeError if not a string
            ValueError if not 'left' or 'right'
        """
        if not isinstance(side, str):
            raise TypeError(f'Side must be a str, not "{type_name(side)}"')
        if side.lower() not in ["left", "right"]:
            raise ValueError(f'Side must be "left" or "right", not {side}')



import logging

from naludaq.communication import DigitalRegisters

from .default import BoardController

LOGGER = logging.getLogger("naludaq.board_controller_aodsoc")


class BoardControllerAodsoc(BoardController):
    """Board controller for AODSOC boards.

    Has a special version of the `read_scalers` function to handle reading from chips
    individually.
    """

    def read_scalar(self, channel: int) -> int:
        """Read the scalar for the given channel"""
        channels_per_chip = self.board.channels // self.board.available_chips
        relative_channel = channel % channels_per_chip
        chip = channel // channels_per_chip
        return self._read_scalar_inner(chip, relative_channel)

    def _read_scalar_inner(self, chip: int, relative_channel: int) -> int:
        """Read the scalar for the given channel.

        This is the inner function that actually reads the scalar.
        It is called by `read_scalar` and should not be called directly.
        """
        name = self.get_scal_name(relative_channel)
        scal = self._read_digital_register(name, chips=chip)
        try:
            scalhigh = self._read_digital_register("scalhigh", chips=chip)
        except (KeyError, AttributeError):
            scalhigh = 0
        shift_amt = DigitalRegisters(self.board).registers[name]["bitwidth"]
        scal += scalhigh << shift_amt

        return scal



"""ASoCv3S Parser, parser for the ASoC with a serial connection.

Use parser to parse the data from the Hardware.
The boards sends a raw bitstream and it needs to be parsed into events.
The parser class is a tool to parse the raw bitstream of 8-bit chunks
into 16 bit words then extracting the 12-bit data, the header info,
channels and window number data.

This data is then returned in a dictionary.
The raw data is preserved.

"""

import logging

import numpy as np

from naludaq.helpers.exceptions import BadDataError
from naludaq.parsers.headers import get_header_parser
from naludaq.parsers.parser import Parser

LOGGER = logging.getLogger(__name__)


class ASoCv3SParser(Parser):
    def __init__(self, params):
        super().__init__(params)
        self.params = params

        self._parse_event_headers = get_header_parser(self.params["model"])

        self.test_amount_windows = 3
        self.samples = self.params.get("samples", 64)
        self.windows = self.params.get("windows", 254)
        self.channels = self.params.get("channels", 4)
        self.num_evt_headers = self.params.get("headers", 3)
        self.num_channel_headers = self.params.get("channel_headers", 1)
        self.num_window_headers = self.params.get("window_headers", 0)
        self.num_window_footers = self.params.get("window_footers", 0)
        self.data_bitmask = self.params.get("data_bitmask", 8191)
        self.chanmask = self.params.get("chanmask", 1023)
        self.windmask = self.params.get("windmask", 255)
        self.chanshift = self.params.get("chanshift", 8)
        self.num_event_footers = self.params.get("footer_words", 1)
        self.last_bits = self.params.get("lastbits", 0)
        self.chan_step_size = self.samples + self.num_channel_headers

        # Generatethe valid bits dictionary
        self.valid_bitlengths = self.create_valid_bitlengths()

    def create_valid_bitlengths(self):
        """ASoCv3S valid bitlentgth combinations

        There are only certain amount of bytes that are valid for the asocv3 serial protocol.

        Returns:
            Dictionary of valid combinations, {bytes: [(channels, windows), (channels, windows)]}
        """
        windows = self.windows
        channels = self.channels
        chan_step_size = self.chan_step_size
        headers = self.num_evt_headers
        win_footers = self.num_window_footers
        num_footers = self.num_event_footers
        bitlengths = {}
        for win in range(1, windows):  # winn never read out 0 windows
            for chan in range(1, channels + 1):
                packlen = (
                    (chan_step_size * chan + win_footers) * win + headers + num_footers
                )
                cur_value = bitlengths.get(packlen, [])
                cur_value.append((chan, win))
                bitlengths[packlen] = cur_value
        return bitlengths

    def get_conflicts(self):
        """Get all bitlengths with ambiguity.

        Args:
            bitlengths (dict):

        """
        potentials = {key: x for key, x in self.valid_bitlengths.items() if len(x) > 1}
        return potentials

    def _validate_input_package(self, in_data):
        """Validate the input data.

        Returns:
            chan (int): Number of channels in data
            wind (int): Number of windows in data

        Raises:
            BadDataException if:
            - there is no data.
            - there is not a en event amount of bytes, since 8-bits are combined to 16-bit words.
            - the data is not divisable by sample_num + window_headers.
            - the data is not a multiple of a combination of windows and channels
        """
        if not isinstance(in_data, dict):
            raise TypeError("raw_data is not a dict it's a %s", type(in_data))

        if in_data["rawdata"] == b"":
            raise BadDataError("No Data Found")

        # Edge case, data doesn't contain valid words.
        if len(in_data["rawdata"]) % 2 != 0:
            raise BadDataError(
                f"Bad package no {in_data.get('pkg_num', 'unknown')}, "
                f"length is not an even amount of words, "
                f"len {len(in_data['rawdata'])}"
            )
        return self.get_chan_wind_from_bitlength(in_data)

    def get_chan_wind_from_bitlength(self, in_data):
        """Get channels and window amoutns from package length."""
        raw_data = np.frombuffer(in_data["rawdata"], dtype=">H")
        raw_data_length = len(raw_data)

        chanwinds = self.valid_bitlengths.get((raw_data_length), [])
        chan = wind = 0

        if not chanwinds:
            raise BadDataError(
                f"Bad package no {in_data.get('pkg_num', 'unknown')}, "
                f"the amount of bytes is not a valid length."
            )

        if len(chanwinds) == 1:
            chan = chanwinds[0][0]
            wind = chanwinds[0][1]

        elif len(chanwinds) > 1:  # potentials.get(rawlen-4, False):
            for chanwind in sorted(chanwinds, key=lambda x: x[0], reverse=True):
                chan = chanwind[0]
                wind = chanwind[1]

                if self.test_chanwinds(raw_data, chan):
                    break

        return chan, wind

    def test_chanwinds(self, raw_data, chans):
        """Check data integrity using the channels
        All windows should be the same in one block of channels.
        If they are not something is wrong, and the check will stop and return False.

        Args:
            raw_data is the the raw data byteshifted and bytes combined to 16-bit words.
            chans is the amount of channel to test for.
        Returns:
            True if test is passed, False on fail.

        """
        step_size = self.chan_step_size * chans + self.num_window_headers
        num_evt_headers = self.num_evt_headers
        last_bits = self.last_bits
        pass_fail = False

        # runs through a predefined number of windows.
        header_steps = range(
            num_evt_headers, step_size * self.test_amount_windows, step_size
        )
        pass_fail_windows = [
            True
        ]  # Converting to set will fail unless the intial value is True.
        for idx in header_steps:
            # step size depends on channel

            windows = []
            window_steps = range(idx, idx + step_size - 1, self.chan_step_size)
            for inner_idx in window_steps:
                pkt = raw_data[inner_idx]
                window = pkt & self.windmask
                window = window >> last_bits
                windows.append(window)
                # check all windownums are the same
            if len(set(windows)) == 1:
                pass_fail_windows.append(True)
            else:
                pass_fail_windows.append(False)
                break

        if len(set(pass_fail_windows)) == 1:
            pass_fail = True

        return pass_fail

    def parse_digital_data(self, in_data):
        """Parse the raw data from the board into events."""
        try:
            # make sure package contains data and there's an even amount of bytes to combine.
            chans, wind = self._validate_input_package(in_data)
        except:
            raise

        channels = self.channels
        num_headers = self.num_evt_headers
        num_footers = self.num_event_footers
        chanmask = self.chanmask
        windmask = self.windmask
        last_bits = self.last_bits
        samples = self.samples
        num_channel_headers = self.num_channel_headers
        num_window_headers = self.num_window_headers
        num_packet = wind
        chanshift = self.chanshift

        data_bitmask = 8191
        raw_data = np.frombuffer(in_data["rawdata"], dtype=">H")
        len(raw_data)

        bitshifted_data = (raw_data & data_bitmask) >> last_bits

        # Create the empty event
        curr_event = {
            "window_labels": [[] for _ in range(channels)],
            "data": [[] for _ in range(channels)],
            "start_window": (raw_data[num_headers] & windmask) >> last_bits,
        }

        # Compute sizes
        data_end = len(bitshifted_data) - num_footers
        len(bitshifted_data) - num_footers - num_headers
        chan_step_size = samples + num_channel_headers
        packet_size = (
            chan_step_size * chans
        ) + num_window_headers  # num_wind_header + samples

        event_matrix = np.reshape(
            bitshifted_data[num_headers:data_end],
            [packet_size, num_packet],
            order="F",
        )
        raw_event_matrix = np.reshape(
            raw_data[num_headers:data_end], [packet_size, num_packet], order="F"
        )

        curr_event["window_timings"] = event_matrix[-1, :]

        for i in range(chans):
            curr_chan = (
                raw_event_matrix[i * chan_step_size, 0] & chanmask
            ) >> chanshift
            curr_event["data"][curr_chan] = (
                event_matrix[1 + i * chan_step_size : (i + 1) * chan_step_size, :]
            ).flatten(order="F")
            curr_event["window_labels"][curr_chan] = (
                raw_event_matrix[i * chan_step_size, :] & windmask
            ) >> last_bits

        return curr_event



"""Base class for the DAC controller, all external DAC controllers should derive from this.

The DAC controller controls the DACs external to the ASICs, different boards are equiped with
one or more DAC chips to set the input Bias on each channel.

The input bias is a DC voltage used to set the base signal at a value above 0, thus making sure
signals with negative values are still positive voltages. Think sinewave, half of the wave is under 0 which
would damage the chip.

Example:
--------

.. code-block:
    dacctrl = get_dac_controller(board)

    # set a sigle channels DAC value:
    dacctrl.set_single_dac(channel, value)

    # Set multiple channels DAC values:
    dacctrl.set_dacs(value, channels,set_mv)

"""
import abc
import logging
from typing import List

from naludaq.controllers.controller import Controller
from naludaq.helpers.exceptions import RegisterFileError

LOGGER = logging.getLogger("naludaq.ext_dac.base")
_REQUIRED_FIELDS = {  # These must be present in the register file.
    "max_mv": int,
    "max_counts": int,
    "channels": dict,
}


class BaseDACController(Controller):
    """Interface for the external/pedestals DAC.

    For the DAC interface to work the board parameters need to be loaded with
    the corresponding DAC parameters, max mV, max counts value set per channel.

    This class needs one channel value per channel in the parameter file or it
    will raise an Error.

    Attributes:
        board (naludaq.board.Board): board object with the corresponding DAC chip params.

    Functions:
        set_single_dac(channel:int, value: int) set the value for a single channel
        set_dacs(value: int, channels: list) sets the same value to multiple channels.

    Raises:
        InvalidRegisterFile is these params are not available in the board.
    """

    def __init__(self, board) -> None:
        """Base class for all DAC controllers.

        Args:
            board (Board): the board object.

        RegisterFileError if the board parameters are missing
                an `ext_dac` field or it is invalid.
        """
        super().__init__(board)
        self.board = board

    @property
    def board(self):
        """Get/Set set board to update DAC chip params.

        Raises:
            RegisterFileError if the board parameters are missing
                an `ext_dac` field or it is invalid.
        """
        return self._board

    @board.setter
    def board(self, board):
        self._validate_ext_dac_params(board.params)
        self.dac_params = board.params["ext_dac"]
        self._board = board

    def set_single_dac(self, channel: int, value: int, set_mv: bool = False):
        """Set the value of a single DAC.

        Args:
            channel(int): channel nummber to set.
            value(int): Value to set DAC to.

        Raises:
            TypeError if channel is not an int
            TypeError if value is not an int
            ValueError if either channel or value is out of bounds.
        """
        try:
            self.set_dacs(value, [channel], set_mv)
        except (TypeError, ValueError):
            raise

    def set_dacs(
        self,
        value: "int | None" = None,
        channels: "list[int] | None" = None,
        set_mv: bool = False,
    ):
        """Sets the values of the external DAC.

        If `channel` is `None`, it'll update all channels with the value given.
        If `value` is `None` it'll use whatever is stored in the board parameters for each channel.

        When changing a DAC, only the specific channels value will change,
        the other channels remain the same (unless stated otherwise).

        To turn a channel off, give it value 0.

        Args:
            value (int): value to set the DAC to. If `None`, values from
                the board object are used instead.
            channels (list): list of channels to set to this value.
            set_mv (bool): whether the value given is in mV.

        Raises:
            TypeError if channels is not a list of integer
            ValueError if channels is not within 0 <= x < < board.channels
            TypeError if the value is not an int.
            ValueError if the value is outside the range 0<=value<max_count
        """
        try:
            if channels is not None:
                self._validate_channels(channels)
            if value is not None:
                self._validate_value(value, set_mv)
        except (TypeError, ValueError):
            raise

        if set_mv:
            value = self._convert_mv2cnt(value)
        if channels is None:
            channels = list(self.board.dac_values.keys())

        original_values = self.board.params["ext_dac"][
            "channels"
        ]  # copy not always needed
        if value is not None:
            original_values = original_values.copy()
            self._set_internal_dac_values(value, channels)

        # An extra check needed in case the user modified the
        # DAC dict directly. ALL DAC values validated out of
        # an abundance of caution.
        try:
            self._validate_internal_dac_values()
            self._write_dacs(channels)
        except Exception as e:
            self.board.params["ext_dac"]["channels"] = original_values
            LOGGER.error(f"Original DAC values restored due to error: {e}")
            raise e

    def set_default_values(self):
        """Sets the dac values for all channels to default values

        Default values are pulled from board.default_params defined
        from the yml
        """
        original_values = self.board.default_params["ext_dac"]["channels"]
        self.board.params["ext_dac"]["channels"] = original_values
        self.set_dacs()

    @abc.abstractmethod
    def _write_dacs(self, channels: List[int]):
        """Implemented by the subclass to handle communication with the
        hardware needed to physically set the DACs.

        Args:
            value (int): the value to set the DACs for the specified channels to.
                If `None`, the DACs are taken from the board object.
            channels (List[int]): the channels to set the DACs for.
        """

    def _validate_ext_dac_params(self, params: dict):
        """Validates the "ext_dac" field on the board params.

        Args:
            params (dict): the params object of the board

        Raises:
            TypeError if the given `params` is not a `dict`.
            RegisterFileError if the external DAC params defined in the
                board configuration are invalid
        """
        if not isinstance(params, dict):
            raise TypeError("Params must be a dict")

        dac_params = params.get("ext_dac", None)
        if dac_params is None:
            raise RegisterFileError("Board does not have an external DAC defined")
        if not isinstance(dac_params, dict):
            raise RegisterFileError("External DAC parameters must be a dict")

        for key, value_type in _REQUIRED_FIELDS.items():
            if key not in dac_params:
                raise RegisterFileError(
                    f'Required field "{key}" is missing from external DAC parameters'
                )
            if not isinstance(dac_params[key], value_type):
                raise RegisterFileError(
                    f'Field "{key}" must be a {value_type.__name__}'
                )

        if not all(isinstance(x, int) for x in dac_params["channels"]):
            raise RegisterFileError("One or more external DAC values is not an int")

    def _validate_value(self, value: int, set_mv: bool = False):
        """Validates the value is within the valid range of counts.

        Args:
            value (int, float): integer value to validate.
                Must be an int if `set_mv` is `False`.
            set_mv (bool): whether to the value given is in mV.

        Raises:
            TypeError if the value is not an int.
            ValueError if th value is outside of the range 0<=value<max_count (12 or 16 bits)
        """
        if not isinstance(set_mv, bool):
            raise TypeError("set_mv must be a bool")
        if not set_mv and not isinstance(value, int):
            raise TypeError("Value in counts must be an int")
        if set_mv and not isinstance(value, (int, float)):
            raise TypeError("Value in mV must be numeric")

        max_value = self.dac_params["max_counts"]
        min_value = self.dac_params.get("min_counts", 0)
        if set_mv:
            max_value = self.dac_params["max_mv"]
            min_value = self.dac_params.get("min_mv", 0)

        if not (min_value <= value <= max_value):
            raise ValueError(
                f"Value is outside range. {min_value} <= value <= {max_value}"
            )

    def _set_internal_dac_values(self, value: int, channels: List[int]):
        """Update the board.dac_values for the selected channels

        Args:
            value (int): the value to set the DACs to
            channels (List[int]): the channels to set the DAC value for.

        Raises:
            TypeError if the arguments are an invalid type.
            ValueError if the value or any channel is out of bounds.
        """
        try:
            self._validate_channels(channels)
            self._validate_value(value)
        except (TypeError, ValueError):
            raise
        for chan in channels:
            self.board.dac_values[chan] = value

    def _validate_internal_dac_values(self):
        """Ensures that all DAC values held in the board object
        are valid. This is important because the user is free
        to modify those values at will, even setting them to
        potentially dangerous values.

        Raises:
            ValueError if _any_ of the DAC values held in the board
                object are out of bounds.
        """
        max_value = self.dac_params["max_counts"]
        for channel, dac_value in self.board.params["ext_dac"]["channels"].items():
            if not 0 <= dac_value <= max_value:
                raise ValueError(
                    f"Cannot set DAC value. Value {dac_value} "
                    f"for channel {channel} exceeds the maximum value"
                )

    def _validate_channels(self, channels: "list[int]"):
        """Make sure the channels are only repeated once and are within range.

        Args:
            channels(list): list of channels to validate

        Raises:
            TypeError if channels is not a list of integers
            ValueError if channels is not within 0 <= x < board.channels
        """
        if not isinstance(channels, list):
            raise TypeError("Channels must be a list of int.")
        if not all(isinstance(chan, int) for chan in channels):
            raise TypeError("Channels must be a list of int.")
        if not all(c in self.board.dac_values.keys() for c in channels):
            raise ValueError("Selected channel numbers are invalid for this board")
        if len(channels) != len(set(channels)):
            raise ValueError("Duplicate channels not allowed")

    def _convert_mv2cnt(self, value):
        """Takes a value in mV and converts it to digital counts based on board model."""
        cnt_hi = self.dac_params["max_counts"]
        cnt_lo = self.dac_params.get("min_counts", 0)
        vref_hi = self.dac_params["max_mv"]
        vref_lo = self.dac_params.get("min_mv", 0)

        cnt = round(value * (cnt_hi - cnt_lo) / (vref_hi - vref_lo))

        return cnt



from typing import List

from naludaq.devices.ad5674r import AD5674R_SPI
from naludaq.controllers.external_dac.base import BaseDACController
from collections import defaultdict


def require_initialized():
    """Check if the SPI connection is initialized."""

    def wrapped_func(func):
        def wrapper(self, *args, **kvargs):
            self._validate_spi_connection_or_raise()
            func(self, *args, **kvargs)

        return wrapper

    return wrapped_func


class DACControllerHDSoCv2(BaseDACController):
    def __init__(self, board):
        """HDSoCv2 DAC controller.

        HDSoCv2 uses Daisy Chained SPI DAC chips AD5674 in two chains
        Due to this configuration the DAC controller works a bit different.
        However the API should be the same as all other DAC controllers.

        Args:
            board: the board object must contain the SPI daisy chains.
        """
        self.board = board

    @property
    def spi(self):
        return self.board.spi_connection

    def _validate_spi_connection_or_raise(self):
        """Validate the SPI connection for the board.

        The spi connection should have two buses for the DACs.
        Each bus should have two devices.
        The two devices should have the channelmap set to the correct values.

        The spi chain should be initialized and the devices should be initialized.

        If the validation fails, an exception will be raised.

        Raises:
            AttributeError: If the validation fails.
        """
        if self.spi is None:
            raise AttributeError("SPI connection is not set")
        if not len(self.spi) >= 2:
            raise AttributeError(
                f"SPI connection should have two buses for the DACs got {len(self.spi)}"
            )
        if len(self.spi[0b01]) != 2 or len(self.spi[0b10]) != 2:
            raise AttributeError("Each bus should have two devices")
        for bus in self.spi:
            for device in bus:
                if not isinstance(device, AD5674R_SPI):
                    raise AttributeError(
                        f"The devices should be AD5674R_SPI objects, got {device}"
                    )
            if not bus._initialized:
                raise AttributeError(
                    f"The bus should be initialized: {bus._initialized}"
                )
        return True

    def _get_device_num(self, channel):
        """Return (device_id, bus_id) for a channel.

        This function will compute the correct bus and device number.
        The devices are connected on the HDSoCv2 eval board in the following way:
        - 0-15: West DAC1
        - 16-31: West DAC0
        - 32-47: East DAC1
        - 48-63: East DAC0

        Args:
            channel: The channel number.

        Returns:
            Tuple of (device_id, bus_id)

        Raises:
        TypeError: If the channel is not an int.
            ValueError: If the channel is not in the range of 0-63.
        """
        bus_num_to_id = {
            0: 0b01,
            1: 0b10,
        }
        validate_channel_or_raise(self.board, channel)
        device_num, _ = divmod(channel, 16)
        bus, dev = divmod(device_num, 2)
        dev = 1 - dev  # invert the device number
        return (bus_num_to_id[bus], dev)

    @require_initialized()
    def _write_dacs(self, channels: List[int]):
        """Write the values from the internal DAC values to the DACs.

        Args:
            channels: List of channels

        """
        dac_values = self.board.dac_values

        chmap = defaultdict(list)
        valmap = defaultdict(list)
        for channel in channels:
            devnum = self._get_device_num(channel)
            chmap[devnum].append(channel)
            valmap[devnum].append(dac_values[channel])

        for (bus_id, dev_id), chans in chmap.items():
            device = self.spi[bus_id][dev_id]
            vals = valmap[devnum]
            if len(chans) == 16 and len(set(vals)) == 1:  # check for broadcast
                device.broadcast(valmap[devnum][0])
            else:
                for i, channel in enumerate(chans):
                    device.set_channel(channel, vals[i])


def validate_channel_or_raise(board, channel):
    if not isinstance(channel, int):
        raise TypeError("Channel must be an int")
    if not (0 <= channel < board.channels):
        raise ValueError(f"Channels must be between 0 and {board.channels - 1}")



"""This module contains controls for selecting which chips are being written/read.
"""
import contextlib

from naludaq.helpers import validations

from .control_registers import ControlRegisters


@contextlib.contextmanager
def select_chips(board, chips: list[int], restore: bool = True):
    """Context manager for selecting which chips are being written/read.

    On context exit, the chips are restored to the previous selection.

    Args:
        board (Board): the board
        chips (list[int]): list of chips to enable for the duration of the context
        restore (bool): whether to restore the previous selection on context exit.
            Can disable this if you are performing many chip-specific operations
            in a row.

    Example:
    ```py
    with select_chips(board, [0, 1]):
        board.connection.send("write something to chips 0 and 1")
    ```
    """
    validations.validate_chip_list_or_raise(chips, board.params)
    select, deselect = select_chips_commands(board, chips)
    _send_command(board, select)
    try:
        yield
    finally:
        if restore:
            _send_command(board, deselect)


def select_chips_commands(board, chips: list[int]) -> tuple[str, str]:
    """Generate a pair of commands to select/restore chips.

    Args:
        board (Board): the board
        chips (list[int]): list of chips to enable

    Returns:
        tuple[str, str]: select, restore commands
    """
    restore = True
    if board.model in ['trbhm', 'dsa-c10-8']:
        restore = False
    prev_chips = selected_chips(board)
    select_cmd = _select_chips_command_factory(board, chips)
    restore_cmd = _select_chips_command_factory(board, prev_chips, restore=restore)
    return select_cmd, restore_cmd


def selected_chips(board) -> list[int]:
    """Get the chips which are currently selected.

    Args:
        board: the board

    Returns:
        list[int]: list of selected chips
    """
    fn = {
        "hiper": _selected_chips_hiper,
        "aodsoc_asoc": _selected_chips_aodsoc,
        "aodsoc_aods": _selected_chips_aodsoc,
        "trbhm": _selected_chips_trbhm,
        "dsa-c10-8": _selected_chips_trbhm,
        "upac96": _selected_chips_upac96,
    }.get(board.model, None)
    result = []
    if fn is not None:
        result = fn(board)
    return result


def wrap_command(
    board, command: str, chips: list[int], restore: bool = True, wait: bool = True
) -> str:
    """Wrap a command in chip selection commands.

    Args:
        board (Board): the board
        command (str): the command to wrap
        chips (list[int]): list of chips to enable
        restore (bool): whether to include a restoration command

    Returns:
        str: wrapped command
    """
    select, restore_cmd = select_chips_commands(board, chips)
    wait_cmd = ""
    if wait:
        wait_cmd = board.params.get("wait", "AE000FFF")
    if not restore:
        restore_cmd = ""

    ret_cmd = f"{select}{wait_cmd}{command}{wait_cmd}{restore_cmd}"
    return ret_cmd


def _select_chips_command_factory(board, chips: list[int], restore: bool = True) -> str:
    """Generate a command to select/deselect chips for the appropriate board.

    If the board is not multichip, an empty string will be returned.

    Args:
        board (Board): the board
        chips (list[int]): list of chips to enable

    Returns:
        str: _description_
    """
    fn = {
        "hiper": _select_chips_command_hiper,
        "aodsoc_asoc": _select_chips_command_aodsoc,
        "aodsoc_aods": _select_chips_command_aodsoc,
        "trbhm": _select_chips_command_trbhm,
        "dsa-c10-8": _select_chips_command_trbhm,
        "upac96": _select_chips_command_upac96,
    }.get(board.model, None)

    command = ""
    if fn is not None:
        command = fn(board, chips, restore)
    return command


def _select_chips_command_hiper(board, chips: list[int], restore: bool = True) -> str:
    """Chip select command generator for aodsoc boards"""
    mask = _build_chip_mask_or_raise(board.params, chips)
    return _generate_commands(
        board,
        {
            "rxout_en": mask,
        },
    )


def _select_chips_command_aodsoc(board, chips: list[int], restore: bool = True) -> str:
    """Chip select command generator for aodsoc boards"""
    mask = _build_chip_mask_or_raise(board.params, chips)
    return _generate_commands(
        board,
        {
            "ard_tx_en": mask,
            "ard_rx_en": mask,
        },
    )


def _select_chips_command_trbhm(board, chips: list[int], restore: bool = True) -> str:
    """Chip select command generator for TRBHM"""
    mask = _build_chip_mask_or_raise(board.params, chips)
    writes = {
        "ard_tx_en": mask,
        "ard_rx_en": mask,
    }
    if restore:
        if len(chips) != 0:
            writes["chip_select"] = chips[0]
    return _generate_commands(board, writes)


def _select_chips_command_upac96(board, chips: list[int], restore: bool = True) -> str:
    """Chip select command generator for upac96"""
    mask = _build_chip_mask_or_raise(board.params, chips)
    writes = {"udc_rxout_enable": mask}
    if len(chips) != 0:
        writes["udc_select"] = chips[0]
    return _generate_commands(board, writes)


def _selected_chips_hiper(board) -> list[int]:
    """Get selected chips for an aodsoc board"""
    mask = board.registers["control_registers"]["rxout_en"]["value"]
    numchips = board.params["num_chips"]
    maxval = 2**numchips - 1
    mask = max(0, min(mask, maxval))
    chiplist = _build_chip_list_or_raise(mask)
    return chiplist


def _selected_chips_aodsoc(board) -> list[int]:
    """Get selected chips for an aodsoc board"""
    mask = board.registers["control_registers"]["ard_rx_en"]["value"]
    return _build_chip_list_or_raise(mask)


def _selected_chips_trbhm(board) -> list[int]:
    """Get selected chips for an trbhm board"""
    mask = board.registers["control_registers"]["ard_rx_en"]["value"]
    return _build_chip_list_or_raise(mask)


def _selected_chips_upac96(board) -> list[int]:
    """Get selected chips for a upac96 board"""
    mask = board.registers["control_registers"]["udc_rxout_enable"]["value"]
    return _build_chip_list_or_raise(mask)


def _build_chip_mask_or_raise(params: dict, chips: list[int]) -> int:
    """Build a chip mask from a list of chips. This function is the inverse of
    `_build_chip_list_or_raise`.

    The mask is a bitfield where each bit represents a chip. The least significant bit
    represents chip 0, the next bit chip 1, and so on.

    Args:
        params (dict): board params, used for validation
        chips (list[int]): list of enabled chips

    Returns:
        int: the chip mask

    Raises:
        ValueError: if the list of chips is invalid
    """
    validations.validate_chip_list_or_raise(chips, params)
    return sum(1 << c for c in chips)


def _build_chip_list_or_raise(chip_mask: int) -> list[int]:
    """Parses a chip mask into a list of chips. This function is the inverse of
    `_build_chip_mask_or_raise`.

    The mask is expected to be a bitfield where each bit represents a chip. The least
    significant bit represents chip 0, the next bit chip 1, and so on.

    Args:
        chip_mask (int): the chip mask

    Returns:
        list[int]: list of enabled chips
    """
    return [
        chip for chip in range(chip_mask.bit_length()) if (chip_mask >> chip) & 1 == 1
    ]


def _send_command(board, command: str):
    """Send a command to the board"""
    from naludaq.backend import BoardIoManager

    if board.using_new_backend:
        BoardIoManager(board).write(command)
    else:
        board.connection.send(command)


def _generate_commands(board, writes: dict[str, int]) -> str:
    cr = ControlRegisters(board)
    return "".join(cr.generate_write(name, value) for name, value in writes.items())



"""Peripherhals controller for UPAC board series

"""
import math
from logging import getLogger

from naludaq.communication import ControlRegisters

from .peripherals_controller import PeripheralsController

logger = getLogger("naludaq.peripherals_upac")


class UpacPeripheralsController(PeripheralsController):
    """Upac specific peripherals controller

    The upac board uses a different system for the peripherals.
    The fpga reads the values off the onboard chips and stores the valus in the register space.
    This changes the readout to use the registers rather than i2c to gather values.

    Registers:
    Power (P1), Current (I1) and Voltage (S1) for PDBIAS rail:
    ltc2992_p1_in
    ltc2992_i1_in
    ltc2992_s1_in

    Temperature Measurement:
    ltc2992_g1_in

    Power (P2), Current (I2), and Voltage (S2) for 1.2v rail:
    ltc2992_p2_in
    ltc2992_i2_in
    ltc2992_s2_in
    """

    def __init__(self, board):
        super().__init__(board)
        self._reader_functions = {
            "temp": self.read_temperature,
            "pdbias_power": self.read_pdbias_power,
            "pdbias_current": self.read_pdbias_current,
            "pdbias_voltage": self.read_pdbias_voltage,
            "1v2_rail_power": self.read_1v2_rail_power,
            "1v2_rail_current": self.read_1v2_rail_current,
            "1v2_rail_voltage": self.read_1v2_rail_voltage,
            "frequencies": self.read_frequencies,
        }

    def read_temperature(self):
        """Read the ambient temperature of the board.

        The voltage for the temperate is bitshifted up 4 bits.
        On this board 1 count is 1 mV making the readout 1:1.
        """
        voltage = self._read_control_register("ltc2992_g1_in") >> 4
        temperature = self._convert_lmt88_voltage_to_temp(voltage / 1000)
        return temperature

    def read_pdbias_power(self):
        """Read the PDBIAS Power from the LTC2992"""

        top = self._read_control_register("ltc2992_p1_in_23_to_16")
        bottom = self._read_control_register("ltc2992_p1_in_15_to_0")

        res = (top << 16) + bottom
        return res

    def read_pdbias_current(self):
        """Read the PDBIAS Current from the LTC2992"""
        return self._read_control_register("ltc2992_i1_in")

    def read_pdbias_voltage(self):
        """Read the PDBIAS Voltage from the LTC2992"""
        # return self._read_control_register('ltc2992_s1_in')
        v_steps = self._read_control_register("ltc2992_s1_in")
        mv = self._voltage_counts_to_mv(v_steps)
        return mv

    def read_1v2_rail_power(self):
        """Read the 1.2V Rail Power from the LTC2992"""

        top = self._read_control_register("ltc2992_p2_in_23_to_16")
        bottom = self._read_control_register("ltc2992_p2_in_15_to_0")
        res = (top << 16) + bottom
        # res = ((top >> 4) << 16) + (bottom >> 4)
        return res

    def read_1v2_rail_current(self):
        """Read the 1.2V Rail Power from the LTC2990"""
        return self._read_control_register("ltc2992_i2_in")

    def read_1v2_rail_voltage(self) -> int:
        """Read the 1.2V Rail Power from the LTC2992

        When operating in 12-bit mode the return is bitshifted up 4-bits and each step is 25mV.
        > LTC2992 Manual p18 section 'ADC Resolution and Conversion Rate'

        Returns:
            readout value in mV.
        """

        v_steps = self._read_control_register("ltc2992_s2_in")
        mv = self._voltage_counts_to_mv(v_steps)
        return mv

    def read_frequencies(self) -> list:
        """Read frquencies for debug purposes.

        Returns: ro_freq, vcdl_freq, mclk_freq
        """

        # RO Monitor clock
        ro_freq = dict.fromkeys(["a", "b", "c", "d"])
        for item in ro_freq.keys():
            msw = self._read_control_register(
                "psec4a_" + item + "_ro_monitor_clock_count_msw"
            )
            lsw = self._read_control_register(
                "psec4a_" + item + "_ro_monitor_clock_count_lsw"
            )
            ro_freq[item] = (msw << 16) + lsw

        # VCDL Monitor clock
        vcdl_freq = dict.fromkeys(["a", "b", "c", "d"])
        for item in vcdl_freq.keys():
            msw = self._read_control_register(
                "psec4a_" + item + "_vcdl_monitor_clock_count_msw"
            )
            lsw = self._read_control_register(
                "psec4a_" + item + "_vcdl_monitor_clock_count_lsw"
            )
            vcdl_freq[item] = (msw << 16) + lsw

        # MCLK Source clock
        mclk_msw = self._read_control_register("psec4a_mclk_source_clock_count_msw")
        mclk_lsw = self._read_control_register("psec4a_mclk_source_clock_count_lsw")
        mclk_freq = (mclk_msw << 16) + mclk_lsw

        return ro_freq, vcdl_freq, mclk_freq

    @staticmethod
    def _voltage_counts_to_mv(v_cnt):
        """Correctly converts the voltage registers to mV from counts.

        When operating in 12-bit mode the return is bitshifted up 4-bits and each step is 25mV.
        > LTC2992 Manual p18 section 'ADC Resolution and Conversion Rate'

        Args:
            v_cnt: coltage in counts

        Returns:
            voltage in mV."""
        step_mv = 25
        bitshift = 4
        mv = (v_cnt >> bitshift) * step_mv
        return mv

    @staticmethod
    def _convert_lmt88_voltage_to_temp(voltage: float) -> int:
        """Converts the return voltage from the LT2990 and LMT88 to temperature in C"""

        temp = -1481.96 + math.sqrt(2.1962e6 + (1.8639 - voltage) / 3.88e-6)
        return temp

    def _read_control_register(self, name: str) -> int:
        """Shorthand to read register value

        Args:
            name: name of the register
        Returns:
            register value
        """
        return ControlRegisters(self.board).read(name)["value"]



import time
from typing import List

from naludaq.communication.control_registers import ControlRegisters

from .base import BaseDACController


class DACControllerHiper(BaseDACController):
    COMMAND_POWER_UP_AND_UPDATE = 0b0011
    COMMAND_POWER_DOWN = 0b0100  # 16-bit data field is ignored

    def __init__(self, board) -> None:
        """DAC controller for the TRBHM board, which uses two LTC2620 chips.

        In addition to supporting setting the DACs for channels 0-7, this controller
        can also set the DACs for both calibration channels and can power down DACs
        for individual channels.

        The LTC2620 is controlled over SPI, which the firmware has an abstraction
        for. Although the firmware supports daisy chaining multiple LTC2620 chips,
        it has a separate chain for each DAC chip, meaning that each chip is at index
        zero in their respective chains.
        """
        super().__init__(board)
        self._chip_mapping = self.dac_params["chip_mapping"]
        self._channel_mapping = self.dac_params["channel_mapping"]
        self._side_mapping = self.dac_params["side_mapping"]

    def _write_dacs(self, channels: List[int]):
        """Writes the internal DAC values for the given channels to the external DAC chips.

        This function will turn the DACs on if they are off (see datasheet p.12)

        Args:
            channels (list[int]): List of channels to write DACs for

        Raises:
            ValueError: if any of the DAC values on the board object are invalid.
        """
        try:
            self._validate_internal_dac_values()
        except ValueError:
            raise

        values = self.board.dac_values
        for channel in channels:
            if channel // 4 not in self.board.params["installed_chips"]:
                continue
            side = self._side_mapping[channel]
            dac_chip = self._chip_mapping[channel]
            dac_channel = self._channel_mapping[channel]
            self._send_command_to_dac(
                self.COMMAND_POWER_UP_AND_UPDATE,
                side,
                dac_chip,
                dac_channel,
                values[channel],
            )

    def _send_command_to_dac(
        self, command: int, side: int, dac_chip: int, dac_channel: int, value: int
    ):
        """Writes a single instruction to a particular DAC chip.

        There are two DAC chips, and the firmware uses the same set of registers
        to communicate with both. Which chip gets updated is determined by the
        'idac_update' register.

        Args:
            command (int): 4-bit command field
            side (int): 0 for left, 1 for right
            dac_chip (int): the dac chip to write to, 0 or 1.
            dac_channel (int): the dac channel to write to. This is NOT the
                ASIC channel number.
            value (int): the 12-bit value to write to the DAC.
        """
        s = {
            0: "l",
            1: "r",
        }[side]

        # 0. make sure the update signal is low (just in case)
        self._write_control_register(f"dac_{s}_update", 0)

        # 1. set the instruction
        self._write_control_register(f"dac_{s}_cmd", command)
        self._write_control_register(f"dac_{s}_asic", dac_chip)
        self._write_control_register(f"dac_{s}_addr", dac_channel)
        self._write_control_register(f"dac_{s}_value", value)

        # 2. update the DAC
        self._write_control_register(f"dac_{s}_update", 1)
        self._write_control_register(f"dac_{s}_update", 0)

        # writing too fast could theoretically cause subsequent instructions to not be received
        time.sleep(0.01)

    def _write_control_register(self, name: str, value: int):
        """Writes a control register."""
        ControlRegisters(self.board).write(name, value)



"""Generates tuning parameters for isel ramp current and cap select.

Uses linear regression by finding the slope of the average channel data
between the highest and lowest value of cap select at different ramp current
values. The slope and y-int is then used to determine what values of ramp
current and cap select will get the channel average to the target value.

The tool will return a dict with the register names as the keys and the values
is a list of values to program per channel.

    Tuning parameters are of the form:
    {
        0 (Channel): {
            isel_ramp_current: (value),
            isel_cap_select: (value),
        }
    }
"""
from naludaq.io.io_manager import load_calibration_data, save_gzip_data

from .udc16 import UDC16ConversionRampOptimizer


def get_conversion_ramp_optimizer(board):
    """Sets the appropriate isel ramp/cap to put the channels in midrange.

    Args:
        board (Board): the board
        channels (list[int]): channels to generate pedestals for. Defaults to all channels.

    Returns:
        An instantiated Conversion ramp optimizer for the given board.

    Raises:
        NotImplementedError if the given board does not support pedestals.
    """

    conversion_ramp_optimizers = {
        "udc16": UDC16ConversionRampOptimizer,
        "upac96": UDC16ConversionRampOptimizer,
    }.get(board.model, None)

    if not conversion_ramp_optimizers or not board.is_feature_enabled(
        "conversion_ramp_optimizer"
    ):
        raise NotImplemented(
            f'Board "{board.model}" does not have support for the conversion ramp optimizer.'
        )
    return conversion_ramp_optimizers(board)


def load_conversion_ramp_file(
    fname: str,
    board,
):
    """Loads conversion ramp tuning parameters from a file, programs
    the parameters to the board, and returns the tuning parameters

    Args:
        fname (str): File name of gzip pickled tuning dictionary
        board (Board): Board to tune conversion ramp
    """
    suggested_vals = load_calibration_data(fname)
    tuner = get_conversion_ramp_optimizer(board)
    tuner._update_isel_for_channels(suggested_vals)
    board.tuning["conversion_ramp"] = suggested_vals
    return suggested_vals


def save_conversion_ramp_file(fname: str, board):
    """Saves conversion ramp tuning parameters from
    board.tuning to a file as a gzip pickle
    Args:
        fname (str): File name to save gzip pickled tuning dictionary
        board (Board): Board to grab conversion ramp tuning parameters
    """
    data = board.tuning.get("conversion_ramp", None)
    if data is None:
        raise KeyError("Board does not have conversion_ramp tuning parameter")
    save_gzip_data(filename=fname, data=data)



def get_clock_controller(board):
    if (
        board.params.get("clock_file", None) is None
        or board.params.get("si5341_address", None) == "None"
    ):
        return None
        # raise NotImplementedError(
        #     "The loaded Register file doesn't have a clock_file specified"
        # )
    from naludaq.controllers.clock.si5341_controller import Si5341Controller

    return Si5341Controller(board)



import abc
import logging
import time
from typing import List

from naludaq.communication import sendI2cCommand
from naludaq.helpers.exceptions import RegisterFileError

from .base import BaseDACController

LOGGER = logging.getLogger("naludaq.ext_dac.i2c_dac")


class I2CDACController(BaseDACController):
    def __init__(self, board, send_delay: float) -> None:
        """ABC DAC controller for DAC chips that communicate through the I2C
        protocol.

        Args:
            board (Board): the board object
            send_delay (float): the time to wait between sequential I2C
                writes in seconds.

        Raises:
            TypeError if the send delay is not a numeric type
            ValueError if the send delay is a negative value
        """
        if not isinstance(send_delay, (int, float)):
            raise TypeError("Send delay must be numeric")
        if send_delay < 0:
            raise TypeError("Send delay cannot be negative")

        super().__init__(board)
        self._send_delay = send_delay
        self._address_mapping = self.dac_params["address_mapping"]
        self._channel_mapping = self.dac_params["channel_mapping"]

    def _validate_ext_dac_params(self, params: dict):
        """Validates the `ext_dac` field in the YAML.

        Args:
            params (dict): the board params.

        Raises:
            RegisterFileError if the ext_dac params is missing
                any requirements of the base class, or if it is
                missing `address_mapping` or `channel_mapping`
                or they hold invalid values/types.
        """
        super()._validate_ext_dac_params(params)
        dac_params = params["ext_dac"]
        chip = dac_params.get("chip", None)
        address_mapping = dac_params.get("address_mapping", None)
        channel_mapping = dac_params.get("channel_mapping", None)
        if address_mapping is None:
            raise RegisterFileError("External DAC params missing address mapping")
        if channel_mapping is None:
            raise RegisterFileError("External DAC params missing channel mapping")
        if chip is None:
            raise RegisterFileError("External DAC params missing chip field")
        if not isinstance(address_mapping, dict):
            raise RegisterFileError("Address mapping must be a dict")
        if not isinstance(channel_mapping, dict):
            raise RegisterFileError("Channel mapping must be a dict")
        if not isinstance(chip, str):
            raise RegisterFileError("Chip must be a str")

        if not all(isinstance(x, int) and x > 0 for x in address_mapping.values()):
            raise RegisterFileError(
                "One or more values in address mapping is not a non-negative int"
            )
        if not all(isinstance(x, int) and x > 0 for x in address_mapping.values()):
            raise RegisterFileError(
                "One or more values in channel mapping is not a non-negative int"
            )

    def _write_dacs(self, channels: List[int]):
        """Writes the values to the DAC for each board channel given.

        The values written come from the board object.

        Args:
            channels (List[int]): list of channels to set the
                DACs for.

        Raises:
            TypeError if the given channels are of the wrong type
            ValueError if any of the internal DAC values on the board
                are out of bounds.
            ConnectionError if the I2C commands could not be sent
        """
        try:
            self._validate_channels(channels)
            self._validate_internal_dac_values()
        except (TypeError, ValueError):
            raise

        rw = 0
        for channel in channels:
            value = self.board.dac_values[channel]
            addr = self._address_mapping[channel]
            words = self._get_write_words(channel, value)
            LOGGER.info("Setting bias DAC ch%s to: %s", channel, value)
            LOGGER.debug(f"Sent DAC Command: {words}")
            if not sendI2cCommand(self.board, (addr << 1) | rw, words):
                raise ConnectionError("Failed to write I2C DAC")
            time.sleep(self._send_delay)

    @abc.abstractmethod
    def _get_write_words(self, channel: int, value: int) -> List[str]:
        """Get words to write to set the DAC value for a channel.

        Args:
            channel (int): the channel to set the DAC value for.
            value (int): the DAC value.

        Returns:
            A list of bytes to send over the I2C bus.
        """



from typing import TypedDict
from collections import defaultdict
import functools

from naludaq.helpers.exceptions import NotInitializedError
from naludaq.devices.spi_bus import SPIBus
from naludaq.devices.spi_device import SPIDevice
from naludaq.backend.managers.io import BoardIoManager

from logging import getLogger

LOGGER = getLogger("naludaq.spichain")


class SPIHeader(TypedDict):
    ncs_hold: bool
    rnw: bool


def initialized(func):
    """Decorator to check if the daisy chain has been initialized before running a method."""

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if self._initialized is False:
            raise NotInitializedError("Daisy chain not initialized")
        return func(self, *args, **kwargs)

    return wrapper


class SPIDaisyChain(SPIBus):
    def __init__(self, board, chain_addr: int):
        """Keeps track of the devices in a SPI daisy chain.

        The daist chain manager adds the command structure to the message to use
        with the Nalu firmware. The top 8-bits of the command are the SPI header
        for the Nalu command structure.

        The daisy chain relies on the devices being added in the correct order.
        The ordering must match the ordering of the devices on the hardware.

        Information will be pushed through the chain until it reaches the desired devices,
        since it's a chain data needs to be send to fill the chain.

        Args:
            chain_addr: The number of the daisy chain. Used by the firmware to select the correct chain.

        """
        self.board = board
        self._chain_address: int = chain_addr

        self._initialized: bool = False

        _data_width = 24  # bit, or 3-byte # TODO: This is generally set by the device.
        self._devices: defaultdict = defaultdict(None)
        self._messages: defaultdict = defaultdict(None)

    def init(self):
        """Initialize the daisy chain.

        This will set the daisy chain enable bit for each device in the chain.
        """

        for i, dev in enumerate(self._devices.values()):
            cmd = dev.init_cmd
            self._messages[i] = cmd

        messages = []
        for i, msg in enumerate(self._messages.values()):
            header = {
                "ncs_hold": False,
                "rnw": False,
            }
            spi_header = self.generate_spi_header(**header)
            spi_message = f"{spi_header}{msg}"

            messages.append(spi_message.upper())

        for message in messages:
            print(f"SENDING MSG ON: {self._chain_address} - {message}")
            self._send_message(message.upper())

        self._reset_msg_buffer()

        self._initialized = True
        # for i, dev in enumerate(self._devices.values()):
        #     cmd = dev.init_cmd
        #     self._messages[i] = cmd

        # self.push_messages()

    def remove(self, device: SPIDevice):
        """Remove device at a specific address."""
        if device in self._devices.values():
            self._devices.remove(device)
            self._reset_msg_buffer()
        else:
            raise AttributeError("Device not in chain")
        self._initialized = False

    def add_message(self, message: str, device_id: int):
        """Add a message to the message queue."""

        self._messages[device_id] = message

    @initialized
    def push_messages(self):
        """Push the messages to the board.

        This will send the messages in the message queue to the board in reverse order.
        If there is no message for a device, a no-op command will be sent.

        The nCS_HOLD bit will be set to True for all devices except the last one.
        """
        if self._initialized is False:
            raise NotInitializedError("Daisy chain not initialized")

        messages = []
        for i, msg in enumerate(self._messages.values()):
            messages.append(self.generate_message(msg, i))

        for message in messages:
            self._send_message(message.upper())

        self._reset_msg_buffer()

    def generate_message(self, message: "str|None", device_id: int) -> str:
        """Generate the message to send to the board.

        Args:
            message: The message to send.
            device_id: The device id.

        Returns:
            The message to send to the board.
        """
        num_devices = len(self._messages)
        header = {
            "ncs_hold": True,
            "rnw": False,
        }
        if message is None:
            message = self[device_id].no_op()
            header["ncs_hold"] = True
        if device_id == num_devices - 1:
            header["ncs_hold"] = False
        spi_header = self.generate_spi_header(**header)
        spi_message = f"{spi_header}{message}"

        return spi_message.upper()

    def _reset_msg_buffer(self):
        """Empties the message buffer."""
        msg = defaultdict(None)
        for k in self._messages.keys():
            msg[k] = None
        self._messages = msg

    def generate_spi_header(self, ncs_hold: bool, rnw: bool) -> str:
        """Generate the SPI header for the Nalu command structure.

        The RnW bit is set to 0 for write operations, 1 for read operations.

        Args:
            ncs_hold: nCS_HOLD bit
            rnw: RnW bit

        Returns:
            SPI header for the Nalu command structure, top 8-bits of command.
        """
        sel = self._chain_address
        SPI_CMD = 0x5
        lb = (ncs_hold << 3) + (rnw << 2) + sel
        cmd = f"{SPI_CMD:1X}{lb:1X}"
        return cmd.upper()

    def _send_message(self, message: str):
        """Send a message to the board.

        Args:
            message: The message to send.
        """
        # LOGGER.debug("SPICHAIN %s is sending message: %s", self._chain_address, message)
        if self.board.using_new_backend:
            BoardIoManager(self.board).write(message)
        else:
            self.board.connection.send(message)

    def __getitem__(self, index) -> SPIDevice:
        return self._devices[index]

    def __setitem__(self, index: int, value: SPIDevice):
        self._devices[index] = value
        self._messages[index] = None
        self._initialized = False

    def __len__(self):
        return len(self._devices)

    def __iter__(self):
        for dev in self._devices.values():
            yield dev



"""Readout settings for the board.

Controls the readout channels and the readout windows.
"""
from collections import defaultdict
from logging import getLogger

from naludaq.communication import DigitalRegisters

from .default import ReadoutController

LOGGER = getLogger("naludaq.readout_controller_hiper")
MAX_READOUT = 65536


class HiperReadoutController(ReadoutController):
    """Special readout controller for the HiPER.

    Overrides the channel-setting logic.
    """

    def __init__(self, board):
        super().__init__(board)

    def set_readout_channels(self, channels_to_read: list):
        """Select channels to readout.

        TRBHM currently doesn't support setting channels individually
        across both ASICs; their 'chansel' is shared.

        Args:
            channels_to_read(list): List of channel numbers to read
                Accepts channels 1..8 for consistent API, but does
                a logical OR between the upper and lower 4 channels.
        """
        self._validate_channels_or_raise(channels_to_read)

        chan_per_chip = self._get_chan_per_chip()
        channel_map = defaultdict(list)

        for ch in channels_to_read:
            chip, chan = divmod(ch, chan_per_chip)
            channel_map[chip].append(chan)

        for chip, chans in channel_map.items():
            excludechanmask = chan_per_chip**2 - 1
            excludechanmask -= sum([1 << x for x in chans])
            DigitalRegisters(self.board, chips=chip).write(
                "excludechannelmask", excludechanmask
            )

    def get_readout_channels(self) -> "list[int]":
        """Get the enabled readout channels as a list[int]."""
        output = []
        chan_per_chip = self._get_chan_per_chip()

        for chip in range(self.board.params["num_chips"]):
            excludemask = self.board.registers["digital_registers"][
                "excludechannelmask"
            ]["value"][chip]
            excludemask = f"{excludemask:016b}"
            output.extend(
                [
                    chip * chan_per_chip + c
                    for c in range(chan_per_chip)
                    if excludemask[c] == "0"
                ]
            )
        return output

    def _get_chan_per_chip(self):
        # Convert channels into a channels per chip
        nchips = self.board.params.get("num_chips", 14)
        nchannels = self.board.params.get("channels", 56)
        return nchannels // nchips



"""

"""
from naludaq.controllers.readout.default import ReadoutController
from naludaq.communication import DigitalRegisters
from naludaq.controllers.board import get_board_controller


class HDSoCv2ReadoutController(ReadoutController):
    def set_readout_channels(self, channels_to_read: list[int]):
        """Select channels to readout.

        Will turn on the chips necessary to read out from the given channels
        (e.g. channels [15, 16, 17] -> chips [0, 1]) and will turn off
        chips that are not being used.

        Args:
            channels_to_read (List[int]): List of channel numbers to read
        """
        self._validate_channels_or_raise(channels_to_read, min_valid=0)
        get_board_controller(self.board).activate_channels = channels_to_read
        self.select_channels(channels_to_read)

    def get_readout_channels(self) -> list:
        """Returns the channels active for readout."""
        return get_board_controller(self.board).active_channels

    def select_channels(self, channels: list[int]):
        """Select channels by masking them.

        Allows individual channel control and also to select channels for readout.

        Args:
            channels: selection to send the following commands to.
        """

        ch_per_group = 8
        for mask_group in range(self.board.params["channels"] // ch_per_group):
            mask_str = f"chanmask{mask_group}"
            ch_range = list(
                range(mask_group * ch_per_group, (mask_group + 1) * ch_per_group)
            )[::-1]
            bin_list = [str(int(channel not in channels)) for channel in ch_range]
            val = int("".join(bin_list), 2)

            DigitalRegisters(self.board).write(mask_str, val)

    def _compute_write_after_trig(self, write_after_trig):
        """On HDSoC write after trigger is equal to windows"""
        return write_after_trig



from . import _common as helpers
from .registers import ReadRegistersABC, WriteRegistersABC


class FpgaRegisters(WriteRegistersABC, ReadRegistersABC):
    def __init__(self, board, register_type: str):
        """Communication with FPGA registers.

        Args:
            board (Board): the board object.
            register_type (str): the type of registers to use.
        """
        super().__init__(board, register_type)
        self._write_cmd = 0xAF
        self._read_cmd = 0xAD
        self._width_cmd = 2
        self._width_addr = 2  # hex char
        self._width_val = 4  # hex char

    # ---------------------------------------------------------------------
    # Write
    # ---------------------------------------------------------------------
    def generate_write(self, name: str, value: int = None) -> str:
        """Generates a command string for writing to a register.

        This function does not change any hardware or software registers.

        Args:
            name (str): the name of the register
            value (int): the value, or None to use the value held in the software register.

        Returns:
            The command string

        Raises:
            ValueError if the address, value or generated command is invalid
            TypeError if the value is an invalid type
        """
        self._validate_name_or_raise(name)
        name = name.lower()
        addr = self._get_addr_from_name_or_raise(name)
        addr_value = self._get_value_from_addr(addr)
        if value is not None:
            addr_value = helpers.substitute_value(
                addr_value, value, self.registers[name]
            )
        return self.generate_write_addr(addr=addr, value=addr_value)

    def generate_write_addr(self, addr: "int|str", value: int = None) -> str:
        """Generates a command string for writing to a register.

        This function does not change any hardware or software registers.

        Args:
            addr (str): the address of the register
            value (int): the value, or None to use the value held in the software register.

        Returns:
            The command string

        Raises:
            ValueError if the address, value or generated command is invalid
            TypeError if the value is an invalid type
        """
        addr = helpers.normalize_address(addr, self._width_addr)
        if value is None:
            value = self._get_value_from_addr(addr)
        self._validate_value_or_raise(value, bitwidth=self._width_val * 4)
        return self._generate_write_atom(addr, value)

    def _generate_write_atom(self, addr: str, value: int) -> str:
        cmd = f"{self._write_cmd:X}"
        cmd_addr = addr.zfill(self._width_addr)
        val = f"{value:X}".zfill(self._width_val)
        command = f"{cmd}{cmd_addr}{val}"
        return command.upper()

    def set(self, name: str, value: int):
        """Set boardparam digital register, do not change the state of the FPGA

        Args:
            name: Register name.
            value: Register value.
        """
        name = name.lower()
        self._validate_name_or_raise(name)
        self._validate_value_or_raise(value, bitwidth=self.registers[name]["bitwidth"])
        self.registers[name]["value"] = value

    def set_addr(self, addr: str, value: int):
        """Update any registers sharing the addr with their part of the value"""
        self._validate_value_or_raise(value, bitwidth=self._width_val * 4)
        for reg in self.registers_on_address(addr.lower()).values():
            val = value >> reg["bitposition"]
            val &= 2 ** reg["bitwidth"] - 1
            reg["value"] = val

    def _get_value_from_addr(self, addr):
        """Generate the value from the hex address.

        An address can contain multiple registers.
        Generate the value depending on the bit positions of all registers on the address.

        Args:
            addr(str): Hex value of the address, 2 hex char long.

        Returns:
            integer value of the combined registers.

        Raises:
            AttributeError if address is not in the registers.
        """
        regs = self.registers_on_address(addr.lower()).values()
        if len(regs) == 0:
            raise AttributeError("Address not found")

        value = 0
        for reg in regs:
            key_value = reg["value"]
            reg["address"]
            key_pos = reg["bitposition"]
            key_width = reg["bitwidth"]

            value += 2**key_pos * (int(key_value) % (2**key_width))
        return value

    # ---------------------------------------------------------------------
    # Read
    # ---------------------------------------------------------------------
    def generate_read_addr(self, addr: "int | str") -> str:
        return self._generate_read_atom(addr)



""" Aquisitions groups of store events.
"""
import time
from collections import deque
from functools import reduce
from typing import Deque, Union

from naludaq.helpers.helper_functions import get_package_versions
from naludaq.models import acq_converters

Event = deque
_NoneType = type(None)
_no_value = object()


class _AcquisitionLikeMeta(type):
    """Metaclass for the `AcquisitionLike` type"""

    def __instancecheck__(self, instance) -> bool:
        return isinstance(instance, (list, deque, Acquisition))


class AcquisitionLike(metaclass=_AcquisitionLikeMeta):
    """Utility class describing behavior common to Acquisitions.

    Examples:
    ```py
    >>> isinstance([], AcquisitionLike)
    True
    >>> isinstance(my_acquisition, AcquisitionLike)
    True
    ```
    """


def _make_property(
    path: tuple, default=_no_value, setter=True, expected_type: type = _no_value
):
    head, tail = path[:-1], path[-1]

    @property
    def getter(self: "Acquisition"):
        value = reduce(dict.__getitem__, head, self.as_dict())
        if default is not _no_value:
            return value.setdefault(tail, default)
        else:
            return value[tail]

    if setter:

        @getter.setter
        def getter(self: "Acquisition", value):
            if expected_type is not _no_value and not isinstance(value, expected_type):
                raise TypeError(
                    f"{tail} must be a {expected_type}, not a {type(value)}"
                )
            reduce(dict.__getitem__, head, self.as_dict())[tail] = value

    return getter


class Acquisition:
    """Events are gathered in Acquistions.

    When you run a capture the events are gathered in an acquisition together with
    the settings used for the capture and the calibration settings used.
    Acquisitions also contains many attributes describing how the
    events were captured, including readout/trigger settings, board registers, etc.

    This class behaves as a wrapper for a `dict` and provides convenient
    accessors for specific entries. The internal `dict` should contain only
    built-in Python types in order to allow for deserialization without the need
    for the `naludaq` package installed. Do not serialize the Acquisition object
    directly!

    Duck-typing is used so that the class looks like both a `deque`
    (to access events) and a `naludaq.board.Board` (to simplify some NaluDAQ operations).

    Args:
        name (str): Acquisition name.
        acq_num (int): Acquisition id number.
        maxlen (int): Maximum number of events that can be held in the
            internal `deque`, or `None` for no maximum.
        data (dict):

    Functions:
        as_dict(): Returns the internal dictionary
        pop(): Remove and return an event.
        popleft(): Remove and return an event on the left side.
        append(event): Add on event on the right side.
        appendleft(event): Add an event on the left side.
        rotate(n): Rotate events by `n` positions right.

    Attributes:
        notes (str): Store any notes for this acquisition.
        info (dict): All information regarding the board, params and triggers
        pedestals
        caldata
        timingcal


    """

    def __init__(
        self, name: str = "", acq_num: int = 0, maxlen: int = None, data: dict = None
    ):
        self._validate_constructor_args(name, acq_num, maxlen, data)
        naludaq_version, naluscope_version = get_package_versions()

        metadata = {
            "acq_num": acq_num,
            "created_at": time.time(),
            "file_version": acq_converters.LATEST_VERSION,
            "info": {
                "connection": {},
                "readings": {},  # current, voltage, temperature
                "readout_settings": {},
                "trigger_settings": {},
            },
            "naludaq_version": naludaq_version,
            "naluscope_version": naluscope_version,
            "name": name,
            "notes": "",
            "settings": {
                "model": "",
                "params": {},
                "registers": {
                    "analog_registers": {},  # not present on UPAC
                    "control_registers": {},
                    "digital_registers": {},  # not present on UPAC
                    "i2c_registers": {},  # not present on UPAC
                },
            },
        }
        self._data = data or {
            "calibration": {
                "pedestals": None,
                "caldata": None,
                "timingcal": None,
            },
            "events": deque(maxlen=maxlen),
            "metadata": metadata,
        }
        if not self._data.get("metadata", None):
            self._data["metadata"] = metadata

    def _validate_constructor_args(
        self, name: str, acq_num: int, maxlen: int, data: dict
    ):
        """Validates arguments passed to the `__init__` method.

        Args:
            See `Acquisition.__init__`.

        Raises:
            ValueError if `data` and any other arguments are non-default.
            TypeError if the given values have the wrong type.
        """
        if data and (name or acq_num or maxlen):
            raise ValueError("When providing a data dict, other args cannot be used")
        elif not isinstance(name, str):
            raise TypeError("name must be a string")
        elif not isinstance(acq_num, int):
            raise TypeError("acq_num must be an int")
        elif not isinstance(maxlen, (int, _NoneType)):
            raise TypeError("maxlen must be an int")
        elif not isinstance(data, (dict, _NoneType)):
            raise TypeError("data must be a dict")

    def as_dict(self) -> dict:
        """Returns the internal dict holding the acquisition data."""
        return self._data

    events: Deque[Event] = _make_property(["events"])

    # ========================= calibration =========================
    pedestals: dict = _make_property(
        ["calibration", "pedestals"], expected_type=(dict, _NoneType)
    )
    caldata: dict = _make_property(
        ["calibration", "caldata"], expected_type=(dict, _NoneType)
    )
    timingcal: list = _make_property(
        ["calibration", "timingcal"], expected_type=(list, _NoneType)
    )

    # ========================= metadata =========================
    metadata: dict = _make_property(["metadata"], expected_type=dict)

    acq_num: int = _make_property(["metadata", "acq_num"], expected_type=int)
    created_at: float = _make_property(["metadata", "created_at"], expected_type=float)
    file_version: str = _make_property(["metadata", "file_version"], expected_type=str)
    naludaq_version: str = _make_property(
        ["metadata", "naludaq_version"], expected_type=str
    )
    naluscope_version: str = _make_property(
        ["metadata", "naluscope_version"], expected_type=str
    )
    name: str = _make_property(["metadata", "name"], expected_type=(str, int))
    notes: str = _make_property(["metadata", "notes"], expected_type=str)

    connection_info: dict = _make_property(
        ["metadata", "info", "connection"], expected_type=(dict, _NoneType)
    )
    readout_settings: dict = _make_property(
        ["metadata", "info", "readout_settings"], expected_type=(dict, _NoneType)
    )
    trigger_settings: dict = _make_property(
        ["metadata", "info", "trigger_settings"], expected_type=(dict, _NoneType)
    )
    readings: dict = _make_property(
        ["metadata", "info", "readings"], expected_type=(dict, _NoneType)
    )

    # ========================= Quack like a Board =========================
    model: str = _make_property(
        ["metadata", "settings", "model"], default="", expected_type=str
    )
    channels: int = _make_property(
        ["metadata", "settings", "params", "channels"], default=None, expected_type=int
    )
    dac_values: dict = _make_property(
        ["metadata", "settings", "params", "ext_dac", "channels"],
        default=None,
        expected_type=dict,
    )
    ext_dac: dict = _make_property(
        ["metadata", "settings", "params", "ext_dac"], default=None, expected_type=dict
    )
    params: dict = _make_property(
        ["metadata", "settings", "params"],
        default=None,
        expected_type=(dict, _NoneType),
    )
    registers: dict = _make_property(
        ["metadata", "settings", "registers"],
        default=None,
        expected_type=(dict, _NoneType),
    )

    # ========================= Quack like a Deque =========================
    def pop(self) -> Event:
        return self.events.pop()

    def popleft(self) -> Event:
        return self.events.popleft()

    def append(self, evt: Event):
        self.events.append(evt)

    def appendleft(self, evt: Event):
        self.events.appendleft(evt)

    def extend(self, evts):
        self.events.extend(evts)

    def extendleft(self, evts):
        self.events.extendleft(evts)

    def clear(self):
        self.events.clear()

    def rotate(self, n: int = 1):
        self.events.rotate(n)

    def remove(self, evt: Event):
        self.events.remove(evt)

    def reverse(self):
        self.events.reverse()

    def __len__(self) -> int:
        return len(self.events)

    def __iter__(self):
        return self.events.__iter__()

    def __getitem__(self, k) -> Event:
        return self.events[k]

    def __setitem__(self, k, v: Event):
        self.events[k] = v

    def __eq__(self, other: Union[list, deque, "Acquisition"]):
        if isinstance(other, (list, deque)):
            return self.events == other
        elif isinstance(other, Acquisition):
            return self.events == other.events
        return False



"""Trigger controller.

The trigger controller is an interface to access and operate the
trigger functionality in the Nalu hardware.
The trigger is active when the chip is in the trigger mode,
by setting BoardController(board).start_readout(*args) where trig
is set to "ext" or "imm".

The trigger is operated by two variables, offset and value.
The value is the ADC threshold value of the trigger.
The offset is an offset from the 0, by setting offset it's possible to
fine-tune the trigger since the needed value is now smaller.

Disable the trigger by setting the value to 0.

Example:

    trig_vals = [1000, 0, 1000, 0]
    board.trigger.values = trig_vals

AUTHOR:
Marcus Luck <marcus@naluscientific.com>

"""
import logging
from enum import Enum

from naludaq.communication import AnalogRegisters, ControlRegisters, DigitalRegisters
from naludaq.controllers.controller import Controller
from naludaq.helpers.semiton import SemitonABC

LOGGER = logging.getLogger(__name__)  # pylint ignore=invalid-name


class TriggerEdge(Enum):
    """Enum for the trigger edge.

    The trigger edge can be set to rising or falling.
    """

    RISING = True
    FALLING = False


class BaseTriggerController(Controller, SemitonABC):
    def __init__(self, board):
        super().__init__(board)
        self.num_banks = 1
        triggers_available = self.params.get("triggers_available", self.board.channels)
        self._trigger_offsets: dict[int, int] = {
            ch: 0 for ch in range(triggers_available)
        }
        self._trigger_values: dict[int, int] = {
            ch: 0 for ch in range(triggers_available)
        }
        edges_available: int = self.params.get("edges_available", 1)
        self._edges: dict[int, TriggerEdge] = {
            ch: TriggerEdge.RISING for ch in range(edges_available)
        }
        # init wbias
        if self.params.get("wbias", None) is None:
            self.params["wbias"] = {ch: 0 for ch in range(self._num_trig_chans)}
        elif isinstance(self.params.get("wbias", None), int):
            self.params["wbias"] = {
                ch: self.params["wbias"] for ch in range(self._num_trig_chans)
            }
        elif isinstance(self.params.get("wbias", None), list):
            self.params["wbias"] = {
                ch: self.params["wbias"][ch] or 0 for ch in range(self._num_trig_chans)
            }
        elif isinstance(self.params.get("wbias", None), dict):
            self.params["wbias"] = {
                ch: self.params["wbias"].get(ch, 0)
                for ch in range(self._num_trig_chans)
            }
        else:
            self.params["wbias"] = {ch: 0 for ch in range(self._num_trig_chans)}
        self._wbias_values: dict[int, int] = {ch: 0 for ch in range(triggers_available)}

    @property
    def bank_size(self) -> int:
        return self.board.channels // self.num_banks

    @property
    def params(self):
        return self.board.params.get("trigger", {})

    @property
    def _num_trig_chans(self):
        return self.board.channels

    @property
    def _max_thresholds(self):
        """Get the maximum threshold value for the board.

        Returns:
            Maximum threshold value.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @property
    def _min_thresholds(self):
        """Get the minimum threshold value for the board.

        Returns:
            Minimum threshold value.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @property
    def references(self) -> "dict[int | str, int]":
        """Get the reference channels for the board.

        Returns:
            Dict of reference channels.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @references.setter
    def references(self, value: "dict[int | str, int]"):
        """Set the reference channels for the board.

        Args:
            value: Dict of reference channels.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @property
    def low_references(self) -> "dict[int | str, int]":
        """Get the reference channels for the board.

        Returns:
            Dict of reference channels.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @low_references.setter
    def low_references(self, value: "dict[int | str, int]"):
        """Set the reference channels for the board.

        Args:
            value: Dict of reference channels.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @property
    def high_references(self) -> "dict[int | str, int]":
        """Get the reference channels for the board.

        Returns:
            Dict of reference channels.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @high_references.setter
    def high_references(self, value: "dict[int | str, int]"):
        """Set the reference channels for the board.

        Args:
            value: Dict of reference channels.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @property
    def wbias(self):
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @wbias.setter
    def wbias(self, value):
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @property
    def offsets(self) -> dict[int, int]:
        """Get the trigger offsets for the board.

        Returns:
            List of trigger offsets.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @offsets.setter
    def offsets(self, value: "dict[int, int] | int | list[int] | None"):
        """Set the trigger offsets for the board.

        Args:
            value: List of trigger offsets.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @property
    def values(self) -> dict[int, int]:
        """Get the trigger values for the board.

        Returns:
            List of trigger values.
        """
        return self.thresholds()

    @values.setter
    def values(self, value: "dict[int, int] | int | list[int] | None"):
        self.thresholds = value

    @property
    def thresholds(self) -> dict[int, int]:
        """Get the trigger thresholds for the board.

        Returns:
            List of trigger thresholds.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @thresholds.setter
    def thresholds(self, value: "dict[int, int] | int | None"):
        """Set the trigger thresholds for the board.

        Args:
            value: List of trigger thresholds.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @property
    def enabled(self) -> dict[int, bool]:
        """Get the trigger enabled per channel.

        Returns:
            List of trigger enabled.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @enabled.setter
    def enabled(self, value: dict[int, bool]):
        """Set the trigger enabled per channel.

        Args:
            value: List of trigger enabled.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @property
    def edge(self) -> "bool | TriggerEdge":
        """Get the trigger edge for the board, opnly available for models with one edge control.

        Returns:
            Trigger edge.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @edge.setter
    def edge(self, value: "bool | TriggerEdge"):
        """Set the trigger edge for the board, opnly available for some models.

        Will set all edges to the same value if board supports multiple channels.

        Args:
            value: Trigger edge.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @property
    def edges(self) -> "dict[int | str, bool | TriggerEdge]":
        """Get the trigger edges for the board.

        Returns:
            List of trigger edges.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @edges.setter
    def edges(self, value: "dict[int | str, bool | TriggerEdge]"):
        """Set the trigger edges for the board.

        Args:
            value: List of trigger edges.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    def write_triggers(
        self, trigger_values: "dict[int, int] | int | list[int] | None" = None
    ):
        """Write the trigger values to the hardware.

        This writes the current set triggers to the hardware.
        By setting the trigger_offsets or the trigger_values
        this function is called automatically and the hardware is updated.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @property
    def tsel(self) -> bool:
        """Get the trigger select for the board.

        Returns:
            Trigger select.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    @tsel.setter
    def tsel(self, value: dict[int, bool]):
        """Set the trigger select for the board.

        Args:
            value: Trigger select.
        """
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )


class TriggerController(BaseTriggerController):
    """Primary tool to setup the triggers for an aquisition.

    Attributes:
        board: The board to update the triggers on.
        trigger_offsets: Offsets as a list, one val per channel.
        trigger_values: The trigger values as a list, one val per channel.

    Functions:
        write_triggers: writes all the trigger values to the physical hardware.
    """

    @property
    def _max_thresholds(self):
        """Get the maximum threshold value for the board.

        Returns:
            Maximum threshold value.
        """
        return self.params.get("max_val", 4095)

    @property
    def _min_thresholds(self):
        """Get the minimum threshold value for the board.

        Returns:
            Minimum threshold value.
        """
        return self.params.get("min_val", 0)

    @property
    def wbias(self) -> dict[int, int]:
        """Get the default wbias for the ASIC."""
        return self.params["wbias"]

    @wbias.setter
    def wbias(self, value: "dict[int, int] | list[int] | int"):
        """Set the wbias for the ASIC.

        The wbias controls the width of the trigger pulse.

        Args:
            value: The wbias value to set.
        """
        if isinstance(value, int):
            self.params["wbias"] = {ch: value for ch in range(self._num_trig_chans)}
        elif isinstance(value, list) and len(value) == self._num_trig_chans:
            self.params["wbias"] = {ch: value for ch, value in enumerate(value)}
        elif isinstance(value, dict) and len(value) == self._num_trig_chans:
            self.params["wbias"] = value
        elif isinstance(value, list) and len(value) != self._num_trig_chans:
            raise ValueError("wbias lists must of length %s", self._num_trig_chans)
        else:
            raise TypeError("Voltage offset must be a single number or a list.")

        self._set_wbiases()

    @property
    def edge(self) -> bool:
        """Get set the trigger edge for all channels."""
        return self._edges.get(0, TriggerEdge.RISING).value

    @edge.setter
    def edge(self, rising: "bool | TriggerEdge"):
        if isinstance(rising, bool):
            rval = TriggerEdge(rising)
        elif isinstance(rising, TriggerEdge):
            rval = rising
        else:
            raise TypeError(
                "rising must be a bool or TriggerEdge, got %s", type(rising)
            )
        self.set_trigger_edge(rval)

    @property
    def offsets(self):
        """Set voltage offset for the trigger.

        Set the trigger offset values, can set one value for all channels or
        a specific value for each channel.

        Updates the value on the hardware if connected.

        Args:
            offset(int): voltage offset for all channels.
            offset(list): voltage offset for the individual channels.

        Returns:
            List of current voltage offsets.

        Raises:
            TypeError if voltage is not a list of length num_chans or a single number.
        """
        return self._trigger_offsets

    @offsets.setter
    def offsets(self, value: "dict[int, int] | list[int] | int") -> None:
        if isinstance(value, int):
            self._trigger_offsets = {ch: value for ch in range(self._num_trig_chans)}

        elif isinstance(value, list):
            if len(value) == self._num_trig_chans:
                self._trigger_offsets = {ch: val for ch, val in enumerate(value)}
            else:
                raise ValueError(
                    f"Trigger offsets must be a list of length {self._num_trig_chans}"
                )
        elif isinstance(value, dict):
            for k, v in value.items():
                self._trigger_offsets[k] = v
        else:
            raise TypeError("Voltage offset must be a single number or a list.")

        self.write_triggers()

    @property
    def values(self) -> dict[int, int]:
        """Trigger threshold for the board.

        Set the trigger threshold values, can set one value for all channels or
        a specific value for each channel.

        Updates the value on the hardware if connected.

        Args:
            trigger_val (list or int): Threshold values for triggers per channel.
                Can take an single number for all channels or one value per channel.

        Returns:
            List of current values for the triggers.

        Raises:
            TypeError if value or key isa of the wrong type.
            ValueError if the list is the wrong length
            ValueError if the value is out of bounds.
            ValueError if the key is out of bounds.
        """
        return self._trigger_values

    @values.setter
    def values(self, trigger_val: "int | dict[int, int] | list[int] | None") -> None:
        conv_val = self._convert_to_dict_or_raise(trigger_val)
        self._validate_thresholds_or_raise(conv_val)

        for k, v in conv_val.items():
            self._trigger_values[k] = v

        self.write_triggers()

    def _convert_to_dict_or_raise(self, value) -> "dict[int, int | bool]":
        """Converts int | dict[int, int|bool] | list[int|bool] to dict[int, int|bool] or raises an error."""
        if isinstance(value, bool):
            conv_val = {ch: value for ch in range(self._num_trig_chans)}
        if isinstance(value, int):
            conv_val = {ch: value for ch in range(self._num_trig_chans)}
        elif isinstance(value, list):
            if len(value) != self._num_trig_chans:
                raise ValueError(
                    f"Trigger values must be a list of length {self._num_trig_chans}"
                )
            conv_val = {ch: val for ch, val in enumerate(value)}
        elif isinstance(value, dict):
            conv_val = value
        else:
            raise TypeError(
                "Trigger values must be: int | bool | dict[int, int|bool] | list[int|bool]"
            )
        return conv_val

    def _validate_thresholds_or_raise(self, value: dict[int, int]) -> None:
        """Validate the trigger thresholds or raise an error."""
        for k, v in value.items():
            if not isinstance(k, int):
                raise KeyError("Key must be an integer")
            if not isinstance(v, int):
                raise TypeError("Values must be an integer")
            if not self._min_thresholds <= v <= self._max_thresholds:
                raise ValueError(
                    f"Value {v} is out of bounds for range {self._min_thresholds} - {self._max_thresholds}"
                )
            if k not in range(self._num_trig_chans):
                raise ValueError(
                    f"Channel {k} is not in range 0:{self._num_trig_chans}"
                )

    def write_triggers(
        self, trigger_values: "dict[int, int] | int | list[int] | None" = None
    ):
        """Write the trigger values to the hardware.

        This writes the current set triggers to the hardware.
        By setting the trigger_offsets or the trigger_values
        this function is called automatically and the hardware is updated.
        """
        self._set_trigger_thresholds(trigger_values)
        self._set_wbiases()
        self._set_trigger_offsets()

        return True

    def _set_trigger_thresholds(self, trigger_values: "dict[int, int] | None" = None):
        """Sets the trigger values and send to hardware.

        Take the trigger values from the board
        """

        trigger_values = trigger_values or self.values
        channels = self.board.channels

        reg = "trigger_threshold_{:02}"
        if self.board.model == "siread":
            reg = "thresh_{:02}"

        for chan in range(channels):
            register = reg.format(chan)
            tval = trigger_values.get(chan, 0)
            self._write_analog_register(register, tval)

    def _set_wbiases(self):
        previous_wbias = self._wbias_values.copy()

        for index in range(self.bank_size):
            wbias_val = 0
            if any(
                x > 0
                for k, x in self.values.items()
                if k
                in range(
                    index * self.num_banks, index * self.num_banks + self.num_banks
                )
            ):
                wbias_val = self.wbias[index]

            register = self._get_wbias_reg_name(index)

            if (
                previous_wbias.get(index, -1) != wbias_val
            ):  # only write if the value has changed
                self._wbias_values[index] = wbias_val
                self._write_analog_register(register, self.wbias[index])

    def _get_wbias_reg_name(self, index):
        return f"wbias_{index:02}"

    def _set_trigger_offsets(self):
        vofs = self.offsets

        channels = self.board.channels
        areg = self.board.registers.get("analog_registers", {})
        if areg.get("offset_00", None) is not None:
            for index in range(channels // self.num_banks):
                register = "offset_" + str(index).zfill(2)
                self._write_analog_register(register, int(vofs[index]))

    def set_trigger_edge(self, rising: "bool | TriggerEdge" = True):
        """Set which signal edge to trigger on.

        Shift between positive going signals (rising) and negative (falling).

        Args:
            rising(bool): If true, trigger on positive going signals, else falling edge.

        Raises:
            TypeError if raises is not a bool.
        """
        if isinstance(rising, bool):
            rising = TriggerEdge(rising)
        if not isinstance(rising, TriggerEdge):
            raise TypeError("rising must be a bool, got %s", type(rising))
        self._edges[0] = rising
        self._write_analog_register("sgn", rising.value)

    def set_trigger_edges(self):
        """Set the trigger edges for individual channels."""
        raise NotImplementedError(
            f"This feature is not available for {self.board.model}"
        )

    def _write_analog_register(
        self, register, value, chips: "int | list[int] | None" = None
    ):
        """wrapper for the Analog register coms module.

        Args:
            register (str): name of the register to update.
            value: The register value to set.

        """
        AnalogRegisters(self.board, chips).write(register, value)

    def _write_control_register(self, register, value):
        """wrapper for the Control register coms module.

        Args:
            register (str): name of the register to update.
            value: The register value to set.

        """
        ControlRegisters(self.board).write(register, value)

    def _write_digital_register(
        self, register, value, chips: "int | list[int] | None" = None
    ):
        """wrapper for the Digital register coms module.

        Args:
            register (str): name of the register to update.
            value: The register value to set.

        """
        DigitalRegisters(self.board, chips).write(register, value)



"""Initializer for the aodsv2 eval board

This initializer will run the init sequence for the aodsv2 eval board.
It contains the board-specific startup sequence.
"""
import logging
import time

from naludaq.board.initializers import Initializers
from naludaq.communication import sendI2cCommand
from naludaq.helpers.decorators import log_function

logger = logging.getLogger("naludaq.init_aodsv2_eval")


class InitAodsv2Eval(Initializers):
    def __init__(self, board):
        """Initializer for AODSv2.

        Args:
            board (Board): the board to initialize.
        """
        super().__init__(board, "aodsv2_eval")
        self.power_sequence = [
            "2v5_en",
        ]

    def run(self) -> bool:
        """Runs the initialization sequence.

        Returns:
            True, always.
        """
        # Initialize FPGA registers
        self._fpga_init()

        # Power on FMC board
        self._power_toggle(True)
        time.sleep(0.25)

        # Program clock
        self._program_clock()

        # Select parallel interface
        self._system_reset()
        self._set_com_mode()

        # Chip-side register startup
        self._analog_startup()
        self._digital_startup()

        # External devices
        self._init_dacs()
        self._init_i2c_devices()

        self.digital_registers.write("waitread", 1)

        return True

    @log_function(logger)
    def _fpga_init(self):
        """Write all FPGA registers to their default values."""
        self.control_registers.write_all()
        self.i2c_registers.write("i2c_en", True)

    @log_function(logger)
    def _power_toggle(self, state):
        """Toggles the power rails defined in `power_sequence` attribute."""
        for register in self.power_sequence:
            self.control_write(register, state)

    @log_function(logger)
    def _system_reset(self):
        """Toggle the sysrst pin high, then low"""
        self.control_registers.write("sysrst", True)
        self.control_registers.write("sysrst", False)

    @log_function(logger)
    def _set_com_mode(self):
        """Select the parallel interface"""
        self.control_registers.write("iomode0", True)
        self.control_registers.write("iomode1", False)

    @log_function(logger)
    def _program_clock(self):
        """Program the clock chip, uses the clockfile in the paramters if it exits."""
        logger.info("Programming clock")
        self.board.clock.program(filename=self.board.params["clock_file"])
        time.sleep(0.25)
        self.control_registers.write("clk_noe", False)

    @log_function(logger)
    def _digital_startup(self):
        """Startup the digital side of the chip by programming all registers."""
        self.digital_registers.write_all()

    @log_function(logger)
    def _analog_startup(self):
        """Start the analog side of the chip by programming all registers."""
        self.analog_registers.write_all()
        self._dll_startup()

    @log_function(logger)
    def _dll_startup(self):
        """Starting the delay line on the analog side.

        Sets and unsets vanbuff to get the dll going and
        changes the vadjp values to ensure proper SST duty cycle once locked.
        """
        self.analog_registers.write("qbuff", 0)
        self.analog_registers.write("vanbuff", 0xB00)
        time.sleep(1)
        self.analog_registers.write("qbuff", 2048)
        self.analog_registers.write("vanbuff", 0)

    @log_function(logger)
    def _init_dacs(self):
        """Write the DAC values using the defaults in the YAML."""
        for channel, value in self._get_dac_values().items():
            self.board.ext_bias.set_single_dac(channel, value)

    def _get_dac_values(self) -> dict:
        """Return a dict with all the {channels: dac_values}"""
        return self.board.params.get("ext_dac", {}).get("channels", {})

    @log_function(logger)
    def _init_i2c_devices(self):
        """Initialize I2C devices on the eval card"""
        sendI2cCommand(self.board, "30", ["05"])  # tempSensor to temp reg
        sendI2cCommand(self.board, "D0", ["00111011"])  # set up current sense ADC



import pytest
from naludaq.board import Board

BOARD_MODELS = ["aodsoc_asoc", "aodsoc_aods"]


class TestTriggerControllerAodsoc:
    @pytest.fixture(autouse=True)
    def setup_board(self):
        self.boards = {}
        for model in BOARD_MODELS:
            self.boards[model] = Board(model)

    @pytest.mark.parametrize("model", BOARD_MODELS)
    def test_set_wbiases(self, mocker, model):
        board = self.boards[model]
        # Mock the _write_analog_registers_per_chip method
        tc_mock = mocker.patch.object(board.trigger, "_write_analog_registers_per_chip")
        mocker.patch.object(
            board.trigger, "values", {0: 1, 1: 0, 2: 1, 3: 0, 4: 1, 5: 0, 6: 1, 7: 0}
        )
        mocker.patch.object(
            board.trigger,
            "wbias",
            {0: 10, 1: 20, 2: 30, 3: 40, 4: 50, 5: 60, 6: 70, 7: 80},
        )
        # Expected wbias values after processing
        expected_wbias = {0: 10, 1: 0, 2: 30, 3: 0, 4: 50, 5: 0, 6: 70, 7: 0}
        board.trigger._set_wbiases()

        # Assert the _write_analog_registers_per_chip method was called with the correct arguments
        tc_mock._write_analog_registers_per_chip.assert_called_once_with(
            "wbias_{:02}", expected_wbias
        )



"""

"""
import time
from logging import getLogger

from naludaq.backend.managers.connection import ConnectionManager
from naludaq.communication.control_registers import ControlRegisters

from . import ConnectionController

LOGGER = getLogger(__name__)


class UPACConnectionController(ConnectionController):
    """Connection controller for UPAC boards

    Uses a different command to change speed in UART connections
    """

    def change_speed(self, sync_int, rate, divider, sync_addr=0x00):
        """
        Attempts to change the baudrate!

        Divider sets it so: (25MHz clock / 16clks/bit)/(divider+1) = baudrate

        NOTE: It is a 100MHz clock now, so the math changes accordingly
          Also it looks like the baud gets quantized on the CPU side at higher rates, so keep that in mind

        Command is DDDDVVVV where V is the divider value (max 256)

        Args:
            sync_int (int): expected integer return when syncing with the board.
            rate (int): valid baudrate
            divider (int): baud divider
        """
        if self.board.using_new_backend:
            return self._change_speed_new(rate, divider)
        if not self.board.connection.is_uart():
            raise ConnectionError

        ControlRegisters(self.board).write("baud_rate_divisor", divider)
        time.sleep(
            0.1
        )  # Important, without wait the baord gets stuck on the next command
        self.board.connection.baud = rate

        # Resyncing the connection.
        for _ in range(3):
            if self.board.connection.resync(sync_int, sync_addr):
                LOGGER.debug(
                    "Baudrate synced successfully, divider: %s, baud: %s", divider, rate
                )
                break
        else:
            LOGGER.debug("Sync failed.")

    def _change_speed_new(self, rate, divider):
        """
        Attempts to change the baudrate!

        Divider sets it so: (25MHz clock / 16clks/bit)/(divider+1) = baudrate

        NOTE: It is a 100MHz clock now, so the math changes accordingly
          Also it looks like the baud gets quantized on the CPU side at higher rates, so keep that in mind

        Command is DDDDVVVV where V is the divider value (max 256)

        Args:
            sync_int (int): expected integer return when syncing with the board.
            rate (int): valid baudrate
            divider (int): baud divider
        """
        self._validate_connection_or_raise()
        ControlRegisters(self.board).write("baud_rate_divisor", divider)
        # Important, without wait the baord gets stuck on the next command
        time.sleep(0.1)
        ConnectionManager(self.board).device.baud_rate = rate

        # Resyncing the connection.
        for _ in range(3):
            if self._resync_new():
                LOGGER.debug(
                    "Baudrate synced successfully, divider: %s, baud: %s", divider, rate
                )
                break
        else:
            LOGGER.debug("Sync failed.")
